#!/usr/bin/env python
# encoding: utf-8
"""
Implementation of Sideways Information Passing graph (builds it from a given ruleset)
"""

import unittest, os, sys, itertools, md5
from FuXi.Horn.PositiveConditions import *
from FuXi.Horn.HornRules import Ruleset
from FuXi.Rete.RuleStore import SetupRuleStore, N3Builtin
from FuXi.DLP import SKOLEMIZED_CLASS_NS
from FuXi.DLP.Negation import ProperSipOrderWithNegation
from rdflib.util import first
from rdflib.Graph import Graph
from rdflib.Collection import Collection
from FuXi.Rete.Util import selective_memoize
from cStringIO import StringIO
from pprint import pprint;
from rdflib import Namespace, Variable, BNode

MAGIC = Namespace('http://doi.acm.org/10.1145/28659.28689#')

def iterCondition(condition):
    if isinstance(condition,Exists):
        return iterCondition(condition.formula)
    else:
        return isinstance(condition,SetOperator) and condition or iter([condition])

def normalizeTerm(uri,sipGraph):
    try:
        return sipGraph.qname(uri).split(':')[-1]
    except:
        return uri.n3()

def RenderSIPCollection(sipGraph,dot=None):
    try:
        from pydot import Node,Edge,Dot
    except:
        import warnings
        warnings.warn("Missing pydot library",ImportWarning)
    if not dot:
        dot = Dot(graph_type='digraph')
        dot.leftNodesLookup = {}                
    nodes = {}
    for N,prop,q in sipGraph.query(
        'SELECT ?N ?prop ?q {  ?prop a magic:SipArc . ?N ?prop ?q . }',
        initNs={u'magic':MAGIC}):

        if MAGIC.BoundHeadPredicate in sipGraph.objects(subject=N,
                                                        predicate=RDF.type):
            NCol = [N]
        else:
            NCol=Collection(sipGraph,N)
            
        if q not in nodes:
            newNode=Node(md5.new(q).hexdigest(),
                         label=normalizeTerm(q,sipGraph),
                         shape='plaintext')                
            nodes[q]=newNode        
            dot.add_node(newNode)
        
        bNode = BNode()
        nodeLabel = ', '.join([normalizeTerm(term,sipGraph) 
                      for term in NCol ])
        edgeLabel = ', '.join([var.n3() 
                           for var in Collection(sipGraph,first(sipGraph.objects(prop,
                                                                    MAGIC.bindings)))])
        markedEdgeLabel = ''
        if nodeLabel in dot.leftNodesLookup:
            bNode,leftNode,markedEdgeLabel = dot.leftNodesLookup[nodeLabel]
#            print "\t",nodeLabel,edgeLabel, markedEdgeLabel,not edgeLabel == markedEdgeLabel
        else:
            leftNode=Node(md5.new(bNode).hexdigest(),label=nodeLabel,shape='plaintext')
            dot.leftNodesLookup[nodeLabel] = (bNode,leftNode,edgeLabel)
            nodes[bNode]=leftNode
            dot.add_node(leftNode)
        
        if not edgeLabel == markedEdgeLabel:
            edge = Edge(leftNode,
                        nodes[q],
                        label=edgeLabel)
            dot.add_edge(edge)
    return dot

class SIPGraphArc(object):
    """
    A sip for r is a labeled graph that satisfies the following conditions:
    1. Each node is either a subset or a member of P(r) or {ph}.
    2. Each arc is of the form N -> q, with label X, where N is a subset of P (r) or {ph}, q is a
    member of P(r), and X is a set of variables, such that
    (i) Each variable of X appears in N.
    (ii) Each member of N is connected to a variable in X.
    (iii) For some argument of q, all its variables appear in X. Further, each variable of X
    appears in an argument of q that satisfies this condition.    
    """
    def __init__(self, left, right, variables, graph=None, headPassing = False):
        self.variables=variables
        self.left = left
        self.right = right
        self.graph = graph is None and Graph() or graph
        self.arc = SKOLEMIZED_CLASS_NS[BNode()]
        self.graph.add((self.arc,RDF.type,MAGIC.SipArc))
        varsCol = Collection(self.graph,BNode())
        [ varsCol.append(i) for i in self.variables ]
        self.graph.add((self.arc,MAGIC.bindings,varsCol.uri))
        if headPassing:
            self.boundHeadPredicate = True
            self.graph.add((self.left,self.arc,self.right))
        else:
            self.boundHeadPredicate = False
            self.graph.add((self.left,self.arc,self.right))
    def __repr__(self):
        """Visual of graph arc"""
        return "%s - (%s) > %s"%(self.left,self.variables,self.right)        
        
def CollectSIPArcVars(left,right,phBoundVars):
    """docstring for CollectSIPArcVars"""
    if isinstance(left,list):
        return set(reduce(lambda x,y:x+y,
                          [ hasattr(t,'isHead') and phBoundVars or GetArgs(t,secondOrder=True) 
                           for t in left])).intersection(GetArgs(right,
                                                                 secondOrder=True))
    else:
        incomingVarsToInclude = phBoundVars and phBoundVars or \
            GetArgs(left,secondOrder=True)
        return set(incomingVarsToInclude).intersection(
                                                    GetArgs(right,
                                                            secondOrder=True))
        
def SetOp(term,value):
    if isinstance(term,N3Builtin):
        term.uri=value
    elif isinstance(term,Uniterm):
        if term.op == RDF.type:
            term.arg[-1]=value
        else:
            term.op=value
    else:
        raise term        

def GetOp(term):
    if isinstance(term,N3Builtin):
        return term.uri
    elif isinstance(term,Uniterm):
        return term.op == RDF.type and term.arg[-1] or term.op
    elif isinstance(term,Exists):
        return GetOp(term.formula)
    else:
        raise term        
        
def GetVariables(term,secondOrder=False):
    for v in GetArgs(term,secondOrder):
        if isinstance(v,Variable):
            yield v
        
def GetArgs(term,secondOrder=False):
    if isinstance(term,N3Builtin):
        return [term.argument,term.result]
    elif isinstance(term,Uniterm):
        args=[]
        if term.op == RDF.type:
            if secondOrder and isinstance(term.arg[-1],(Variable, BNode)):
                args.extend(term.arg)
            else:
                args.append(term.arg[0])
        elif isinstance(term.op,(Variable, BNode)):
            args.append(term.op)
            args.extend(term.arg)
        else:
            args.extend(term.arg)
        return args
    elif isinstance(term,Exists):
        return GetArgs(term.formula, secondOrder)
    else:
        raise term        
        
def IncomingSIPArcs(sip,predOcc):
    """docstring for IncomingSIPArcs"""
    for s,p,o in sip.triples((None,None,predOcc)): 
        if (p,RDF.type,MAGIC.SipArc) in sip:
            if (s,RDF.type,MAGIC.BoundHeadPredicate) in sip:
                yield [s],Collection(sip,first(sip.objects(p,MAGIC.bindings)))
            else:
                yield Collection(sip,s),Collection(sip,first(sip.objects(p,MAGIC.bindings)))
        
def validSip(sipGraph):
    if not len(sipGraph): return False
    for arc in sipGraph.query(
         "SELECT ?arc { ?arc m:bindings ?bindings OPTIONAL { ?bindings rdf:first ?val } FILTER(!BOUND(?val)) }",
         initNs={'m':MAGIC}):
        return False
    return True

def getOccurrenceId(uniterm,lookup={}):
    pO = URIRef(GetOp(uniterm)+'_'+'_'.join(GetArgs(uniterm)))
    lookup[pO]=GetOp(uniterm)
    return pO
        
def findFullSip((rt,vars),right):
    if not vars:
        if len(rt)==1:
            vars=GetArgs(rt[0],secondOrder=True)
        else:
            vars=reduce(lambda l,r: [i for i in GetArgs(l,secondOrder=True)+GetArgs(r,secondOrder=True) 
                                                if isinstance(i,(Variable,BNode))],rt)
    if len(right)==1:
        if set(GetArgs(right[0],secondOrder=True)).intersection(vars):#len(dq)==1:
            #Valid End of recursion, return full SIP order
            yield rt+right  
    else: 
        #for every possible combination of left and right, trigger recursive call
        for item in right:
            _vars = set([v for v in GetArgs(item,secondOrder=True) if isinstance(v,(Variable,BNode))])
            _inVars = set([v for v in vars])
            if _vars.intersection(vars):
                #There is an incoming arc, continue processing inductively on
                #the rest of right
                _inVars.update(_vars.difference(vars))
                for sipOrder in findFullSip((rt+[item],_inVars),
                                            [i for i in right if i != item]):
                    yield sipOrder
                    
class InvalidSIPException(Exception):
    def __init__(self,msg=None): 
        super(InvalidSIPException, self).__init__(msg)
        
def BuildNaturalSIP(clause,derivedPreds,adornedHead, ignoreUnboundDPreds = False):
    """
    Natural SIP:
    
    Informally, for a rule of a program, a sip represents a
    decision about the order in which the predicates of the rule will be evaluated, and how values
    for variables are passed from predicates to other predicates during evaluation
    
    >>> ruleStore,ruleGraph=SetupRuleStore(StringIO(PROGRAM2))
    >>> ruleStore._finalize()
    >>> fg=Graph().parse(StringIO(PROGRAM2),format='n3')
    >>> rs=Ruleset(n3Rules=ruleGraph.store.rules,nsMapping=ruleGraph.store.nsMgr)
    >>> for rule in rs: print rule
    Forall ?Y ?X ( ex:sg(?X ?Y) :- ex:flat(?X ?Y) )
    Forall ?Y ?Z4 ?X ?Z1 ?Z2 ?Z3 ( ex:sg(?X ?Y) :- And( ex:up(?X ?Z1) ex:sg(?Z1 ?Z2) ex:flat(?Z2 ?Z3) ex:sg(?Z3 ?Z4) ex:down(?Z4 ?Y) ) )
    >>> sip=BuildNaturalSIP(list(rs)[-1])
    >>> for N,x in IncomingSIPArcs(sip,MAGIC.sg): print N.n3(),x.n3()
    ( <http://doi.acm.org/10.1145/28659.28689#up> <http://doi.acm.org/10.1145/28659.28689#sg> <http://doi.acm.org/10.1145/28659.28689#flat> ) ( ?Z3 )
    ( <http://doi.acm.org/10.1145/28659.28689#up> <http://doi.acm.org/10.1145/28659.28689#sg> ) ( ?Z1 )
    
    >>> sip=BuildNaturalSIP(list(rs)[-1],[MAGIC.sg])
    >>> list(sip.query('SELECT ?q {  ?prop a magic:SipArc . [] ?prop ?q . }',initNs={u'magic':MAGIC}))
    [rdflib.URIRef('http://doi.acm.org/10.1145/28659.28689#sg'), rdflib.URIRef('http://doi.acm.org/10.1145/28659.28689#sg')]
    """
    from FuXi.Rete.Magic import AdornedUniTerm
    occurLookup={}
    boundHead=isinstance(adornedHead,AdornedUniTerm) and 'b' in adornedHead.adornment
    phBoundVars = list(adornedHead.getDistinguishedVariables(varsOnly=True))
    assert isinstance(clause.head,Uniterm),"Only one literal in the head!"
    def collectSip(left,right):
        if isinstance(left,list):
            vars=CollectSIPArcVars(left,right,phBoundVars)
            if not vars and ignoreUnboundDPreds:
                raise InvalidSIPException("No bound variables for %s"%right)            
            leftList=Collection(sipGraph,None)
            left=list(set(left))            
            [leftList.append(i) for i in [GetOp(ii) for ii in left]]
            left.append(right)   
            arc=SIPGraphArc(leftList.uri,getOccurrenceId(right,occurLookup),vars,sipGraph)
            return left
        else:
            left.isHead = True
            vars=CollectSIPArcVars(left,right,phBoundVars)
            if not vars and ignoreUnboundDPreds:
                raise InvalidSIPException("No bound variables for %s"%right)            
            ph=GetOp(left)
            q=getOccurrenceId(right,occurLookup)
            if boundHead:
                arc=SIPGraphArc(ph,q,vars,sipGraph,headPassing=boundHead)
                sipGraph.add((ph,RDF.type,MAGIC.BoundHeadPredicate))
                rt=[left,right]
            else:
                rt=[right]
        return rt
    sipGraph=Graph()  
    if isinstance(clause.body,And):
        if ignoreUnboundDPreds:
            foundSip = False
            sips = findFullSip(([clause.head],None), clause.body)
            while not foundSip:
                sip = sips.next()
                try:
                    reduce(collectSip,
                           iterCondition(And(sip)))
                    foundSip = True
                    bodyOrder = sip                           
                except InvalidSIPException:
                    foundSip = False
        else:
            if first(itertools.ifilter(lambda i:isinstance(i,Uniterm) and i.naf or False,
                                       clause.body)):
                #There are negative literals in body, ensure
                #the given sip order puts negated literals at the end
                bodyOrder=first(
                        itertools.ifilter(ProperSipOrderWithNegation,
                                          findFullSip(([clause.head],None), 
                                                        clause.body)))
            else:
                bodyOrder=first(findFullSip(([clause.head],None), clause.body))
            assert bodyOrder,"Couldn't find a valid SIP for %s"%clause
            reduce(collectSip,
                   iterCondition(And(bodyOrder)))
        sipGraph.sipOrder = And(bodyOrder[1:])
        #assert validSip(sipGraph),sipGraph.serialize(format='n3')
    else:
        if boundHead:
            reduce(collectSip,itertools.chain(iterCondition(clause.head),
                                              iterCondition(clause.body)))
        sipGraph.sipOrder = clause.body
    
    if derivedPreds:
        # We therefore generalize our notation to allow
        # more succint representation of sips, in which only arcs entering 
        # derived predicates are represented.
        arcsToRemove=[]
        collectionsToClear=[]
        for N,prop,q in sipGraph.query(
            'SELECT ?N ?prop ?q {  ?prop a magic:SipArc . ?N ?prop ?q . }',
            initNs={u'magic':MAGIC}):
            if occurLookup[q] not in derivedPreds:
                arcsToRemove.extend([(N,prop,q),(prop,None,None)])
                collectionsToClear.append(Collection(sipGraph,N))
                #clear bindings collection as well
                bindingsColBNode=first(sipGraph.objects(prop,MAGIC.bindings))
                collectionsToClear.append(Collection(sipGraph,bindingsColBNode))
        for removeSts in arcsToRemove:
            sipGraph.remove(removeSts)
        for col in collectionsToClear:
            col.clear()
    return sipGraph

def SIPRepresentation(sipGraph):
    rt=[]
    for N,prop,q in sipGraph.query(
        'SELECT ?N ?prop ?q {  ?prop a magic:SipArc . ?N ?prop ?q . }',
        initNs={u'magic':MAGIC}):
        if MAGIC.BoundHeadPredicate in sipGraph.objects(subject=N,predicate=RDF.type):
            NCol = [N]
        else:
            NCol=Collection(sipGraph,N)
        rt.append( "{ %s } -> %s %s"%(
          ', '.join([normalizeTerm(term,sipGraph) 
                      for term in NCol ]),
          ', '.join([var.n3() 
                      for var in Collection(sipGraph,first(sipGraph.objects(prop,
                                                                            MAGIC.bindings)))]),
          normalizeTerm(q,sipGraph)
                              ))
    return rt
def test():
    import doctest
    doctest.testmod()

if __name__ == '__main__':
    test()


#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Implements a Sip-Strategy formed from a basic graph pattern and a RIF-Core ruleset
as a series of top-down derived SPARQL evaluations against a fact graph, 
generating a walk through the proof space in the process.

Native Prolog-like Python implementation for RIF-Core, OWL 2, and SPARQL.  
"""

import itertools, copy, md5, pickle
from FuXi.Rete.AlphaNode import ReteToken, AlphaNode
from FuXi.Horn.HornRules import Clause, Ruleset, Rule, HornFromN3
from FuXi.Rete.RuleStore import *
from FuXi.Rete.Magic import AdornLiteral
from FuXi.Horn.PositiveConditions import *
from FuXi.Rete.Proof import *
from FuXi.Rete.Util import selective_memoize, lazyGeneratorPeek
from rdflib.Graph import ReadOnlyGraphAggregate
from rdflib import URIRef, RDF, Namespace, Variable
from rdflib.util import first
from rdflib.syntax.xml_names import split_uri
from FuXi.Rete.SidewaysInformationPassing import *
from FuXi.SPARQL import EDBQuery, normalizeBindingsAndQuery

def PrepareSipCollection(adornedRuleset):
    """
    Takes adorned ruleset and returns an RDF dataset
    formed from the sips associated with each adorned
    rule as named graphs.  Also returns a mapping from
    the head predicates of each rule to the rules that match
    it - for efficient retrieval later
    """
    headToRule = {}
    graphs = []
    secondOrderRules = set()
    for rule in adornedRuleset:
        ruleHead = GetOp(rule.formula.head)
        if isinstance(ruleHead,Variable):
            #We store second order rules (i.e., rules whose head is a 
            #predicate occurrence whose predicate symbol is a variable) aside
            secondOrderRules.add(rule)
        headToRule.setdefault(ruleHead,set()).add(rule)
        graphs.append(rule.sip)
    #Second order rules are mapped from a None key (in order to indicate they are wildcards)
    headToRule[None]=secondOrderRules
    if not graphs:
        return
    graph = ReadOnlyGraphAggregate(graphs)
    graph.headToRule = headToRule
    return graph

def getBindingsFromLiteral(groundTuple,ungroundLiteral):
    """
    Takes a ground fact and a query literal and returns
    the mappings from variables in the query literal
    to terms in the ground fact
    """
    ungroundTuple = ungroundLiteral.toRDFTuple()
    return ImmutableDict([(term,groundTuple[idx]) 
                  for idx,term in enumerate(ungroundTuple)
                      if isinstance(term,Variable) and
                         not isinstance(groundTuple[idx],Variable)])

def tripleToTriplePattern(graph,term):
    if isinstance(term,N3Builtin):
        template = graph.templateMap[term.uri]
        return "FILTER(%s)"%(template%(term.argument.n3(),
                                       term.result.n3()))
    else:
        return "%s %s %s"%tuple([renderTerm(graph,term) 
                                    for term in term.toRDFTuple()])
@selective_memoize([0])
def normalizeUri(rdfTerm,revNsMap):
    """
    Takes an RDF Term and 'normalizes' it into a QName (using the registered prefix)
    or (unlike compute_qname) the Notation 3 form for URIs: <...URI...> 
    """
    try:
        namespace, name = split_uri(rdfTerm)
        namespace = URIRef(namespace)
    except:
        if isinstance(rdfTerm,Variable):
            return "?%s"%rdfTerm
        else:
            return "<%s>"%rdfTerm
    prefix = revNsMap.get(namespace)
    if prefix is None and isinstance(rdfTerm,Variable):
        return "?%s"%rdfTerm
    elif prefix is None:
        return "<%s>"%rdfTerm
    else:
        qNameParts = compute_qname(rdfTerm,revNsMap)         
        return ':'.join([qNameParts[0],qNameParts[-1]])    

@selective_memoize([0])
def compute_qname(uri,revNsMap):
    namespace, name = split_uri(uri)
    namespace = URIRef(namespace)
    prefix = revNsMap.get(namespace)
    if prefix is None:
        prefix = "_%s" % len(revNsMap)
        revNsMap[namespace]=prefix
    return (prefix, namespace, name)

def renderTerm(graph,term):
    if term == RDF.type:
        return ' a '
    elif isinstance(term,URIRef):
        qname = normalizeUri(term,hasattr(graph,'revNsMap') and graph.revNsMap or \
                             dict([(u,p) for p,u in graph.namespaces()]))
        return qname[0] == '_' and u"<%s>"%term or qname
    else:
        try:
            return isinstance(term,BNode) and term.n3() or graph.qname(term)
        except:
            return term.n3()

def RDFTuplesToSPARQL(conjunct, 
                      edb, 
                      isGround=False, 
                      vars=[],
                      symmAtomicInclusion=False):
    """
    Takes a conjunction of Horn literals and returns the 
    corresponding SPARQL query 
    """
    queryType = isGround and "ASK" or "SELECT %s"%(' '.join([v.n3() 
                                                             for v in vars]))
    queryShell = len(conjunct)>1 and "%s {\n%s\n}" or "%s { %s }"
    if symmAtomicInclusion:
        if vars:
            var = vars.pop()
            prefix = "%s a ?KIND"%var.n3()
        else:
            
            prefix = "%s a ?KIND"%first([lit.arg[0].n3() for lit in conjunct])
        subquery = queryShell%(queryType,
                               "%s\nFILTER(%s)"%(
                             prefix,
                             ' ||\n'.join([
                               '?KIND = %s'%edb.qname(GetOp(lit)) 
                                    for lit in conjunct])))        
    else: 
        subquery = queryShell%(queryType,' .\n'.join(['\t'+tripleToTriplePattern(
                                                              edb,
                                                              lit) 
                                  for lit in conjunct ]))
    return subquery
    
def lazyCollapseBooleanProofs(left,right):
    """
    Function for reduce that (lazily) performs
    boolean conjunction operator on a list
    of 2-tuples, a boolean value and some object
    . The boolean conjunction is applied on the
    first item in each 2-tuple
    """
    (leftBool,leftNode)   = left
    (rightBool,rightNode) = right
    if not leftBool:  
        return False, None
    else: 
        return (leftBool and rightBool) and (True,rightNode) or (False,None)
                
def literalIsGround(literal):
    """
    Whether or not the given literal has
    any variables for terms
    """
    return not [term for term in GetArgs(literal,
                                         secondOrder=True)
                                         if isinstance(term,Variable) ]
    
def mergeMappings1To2(mapping1,mapping2,makeImmutable=False):
    """
    Mapping merge.  A 'safe' update (i.e., if the key
    exists and the value is different, raise an exception)
    An immutable mapping can be returned if requested
    """
    newMap = {}
    for k,v in mapping1.items():
        val2 = mapping2.get(k)        
        if val2:
            assert v == val2,"Failure merging %s to %s"%(mapping1,mapping2) 
            continue
        else: 
            newMap[k] = mapping1[k]
    newMap.update(mapping2)
    return makeImmutable and MakeImmutableDict(newMap) or newMap
         
class RuleFailure(Exception): 
    def __init__(self, msg):
        self.msg = msg
    def __repr__(self):
        return "RuleFailure: %"%self.msg 

class parameterizedPredicate:
    def __init__(self, externalVar):
        self.externalVar = externalVar
    def __call__(self, f):
        def _func(item):
            return f(item,self.externalVar)
        return _func

def invokeRule(priorAnswers,
               bodyLiteralIterator,
               sip,
               otherargs,
               priorBooleanGoalSuccess=False,
               step = None,
               debug = False,
               buildProof = False):
    """
    Continue invokation of rule using (given) prior answers and list of remaining
    body literals (& rule sip).  If prior answers is a list, computation is split 
    disjunctively
    
    [..] By combining the answers to all these subqueries, we generate
    answers for the original query involving the rule head
    
    Can also takes a PML step and updates it as it navigates the top-down proof
    tree (passing it on and updating it where necessary)

    """
    assert not buildProof or step is not None 
    proofLevel, memoizeMemory, sipCollection, factGraph, derivedPreds, processedRules = otherargs
    remainingBodyList = [i for i in bodyLiteralIterator]
    lazyGenerator = lazyGeneratorPeek(priorAnswers,2)
    if lazyGenerator.successful: 
        #There are multiple answers in this step, we need to call invokeRule
        #recursively for each answer, returning the first positive attempt
        success = False
        rt = None
        _step = None
        ansNo = 0
        for priorAns in lazyGenerator:
            ansNo += 1
            try:
                if buildProof:
                    newStep = InferenceStep(step.parent,
                                            step.rule,
                                            source=step.source)
                    newStep.antecedents = [ant for ant in step.antecedents]
                else:
                    newStep = None
                for rt,_step in\
                   invokeRule([priorAns],
                              iter([i for i in remainingBodyList]),
                              sip,
                              otherargs,
                              priorBooleanGoalSuccess,
                              newStep,
                              debug = debug,
                              buildProof = buildProof):
                    if rt:
                        yield rt,_step
            except RuleFailure, e: pass
        if not success:
            #None of prior answers were successful
            #indicate termination of rule processing
            raise RuleFailure(
              "Unable to solve either of %s against remainder of rule: %s"%(
                    ansNo,remainingBodyList))
#            yield False,_InferenceStep(step.parent,step.rule,source=step.source)
    else:
        lazyGenerator = lazyGeneratorPeek(lazyGenerator)
        projectedBindings = lazyGenerator.successful and first(lazyGenerator) or {}

        #First we check if we can combine a large group of subsequent body literals
        #into a single query        
        #if we have a template map then we use it to further
        #distinguish which builtins can be solved via 
        #cumulative SPARQl query - else we solve
        #builtins one at a time
        def sparqlResolvable(literal):
            if isinstance(literal,Uniterm):
                return not literal.naf and GetOp(literal) not in derivedPreds
            else:
                return isinstance(literal,N3Builtin) and \
                       literal.uri in factGraph.templateMap
        def sparqlResolvableNoTemplates(literal):
            if isinstance(literal,Uniterm):
                return not literal.naf and GetOp(literal) not in derivedPreds
            else:
                return False
                   
        conjGroundLiterals = list(itertools.takewhile(
                          hasattr(factGraph,'templateMap') and sparqlResolvable or \
                          sparqlResolvableNoTemplates,
                          remainingBodyList))
        bodyLiteralIterator = iter(remainingBodyList)
        if len(conjGroundLiterals)>1:
            #If there are literals to combine *and* a mapping from rule
            #builtins to SPARQL FILTER templates ..
            basePredicateVars = set(
                    reduce(lambda x,y:x+y,
                           map(lambda arg:list(GetVariables(arg,secondOrder=True)),
                               conjGroundLiterals)))
            if projectedBindings: 
                openVars = basePredicateVars.intersection(projectedBindings)
            else:
                #We don't have any given bindings, so we need to treat
                #the body as an open query
                openVars = basePredicateVars
            queryConj = EDBQuery([copy.deepcopy(lit) for lit in conjGroundLiterals],
                                    factGraph,
                                    openVars,
                                    projectedBindings)
            query,answers = queryConj.evaluate(debug)
            if isinstance(answers,bool):
                combinedAnswers = {}
                rtCheck = answers
            else:
                if projectedBindings:
                    combinedAnswers = (mergeMappings1To2(ans,
                                                         projectedBindings,
                                                         makeImmutable=True) for ans in answers )
                else:
                    combinedAnswers = ( MakeImmutableDict(ans) for ans in answers )
                combinedAnsLazyGenerator = lazyGeneratorPeek(combinedAnswers)
                rtCheck = combinedAnsLazyGenerator.successful
            if not rtCheck:
                raise RuleFailure("No answers for combined SPARQL query: %s"%query)
            else:
                #We have solved the previous N body literals with a single 
                #conjunctive query, now we need to make each of the literals
                #an antecedent to a 'query' step.
                if buildProof:
                    queryStep = InferenceStep(None,source='some RDF graph')
                    queryStep.groundQuery = subquery
                    queryStep.bindings = {}#combinedAnswers[-1]
                    queryHash = URIRef("tag:info@fuxi.googlecode.com:Queries#"+md5.new(
                                                subquery).hexdigest())
                    queryStep.identifier = queryHash
                    for subGoal in conjGroundLiterals:
                        subNs=NodeSet(subGoal.toRDFTuple(),
                                   identifier=BNode())
                        subNs.steps.append(queryStep)
                        step.antecedents.append(subNs)
                        queryStep.parent = subNs
                for rt,_step in invokeRule(
                               isinstance(answers,bool) and [projectedBindings] or combinedAnsLazyGenerator,
                               iter(remainingBodyList[len(conjGroundLiterals):]),
                               sip,
                               otherargs,
                               isinstance(answers,bool),
                               step,
                               debug = debug,
                               buildProof = buildProof):
                    yield rt,_step
            
                    
        else:
            #Continue processing rule body condition
            #one literal at a time
            try:
                bodyLiteral = bodyLiteralIterator.next()
                
                #if a N3 builtin, execute it using given bindings for boolean answer
                #builtins are moved to end of rule when evaluating rules via sip
                if isinstance(bodyLiteral,N3Builtin):
                    lhs = bodyLiteral.argument
                    rhs = bodyLiteral.result
                    lhs = isinstance(lhs,Variable) and projectedBindings[lhs] or lhs
                    rhs = isinstance(rhs,Variable) and projectedBindings[rhs] or rhs
                    assert lhs is not None and rhs is not None
                    if bodyLiteral.func(lhs,rhs):
                        if debug:
                            print >> sys.stderr, "Invoked %s(%s,%s) -> True"%(
                                             bodyLiteral.uri,
                                             lhs,
                                             rhs)
                        #positive answer means we can continue processing the rule body
                        if buildProof:
                             ns=NodeSet(bodyLiteral.toRDFTuple(),
                                        identifier=BNode())
                             step.antecedents.append(ns)
                        for rt,_step in invokeRule(
                                           [projectedBindings],
                                           bodyLiteralIterator,
                                           sip,
                                           otherargs,
                                           step,
                                           priorBooleanGoalSuccess,
                                           debug = debug,
                                           buildProof = buildProof):
                            yield rt,_step
                    else:
                        if debug:
                            print >> sys.stderr, "Successfully invoked %s(%s,%s) -> False"%(
                                             bodyLiteral.uri,
                                             lhs,
                                             rhs)                
                        raise RuleFailure("Failed builtin invokation %s(%s,%s)"%
                                          (bodyLiteral.uri,
                                           lhs,
                                           rhs))
                else:
                    #For every body literal, subqueries are generated according to the sip      
                    sipArcPred = URIRef(GetOp(bodyLiteral)+'_'+'_'.join(GetArgs(bodyLiteral)))
                    assert len(list(IncomingSIPArcs(sip,sipArcPred)))<2
                    subquery = copy.deepcopy(bodyLiteral)
                    subquery.ground(projectedBindings)
                    for N,x in IncomingSIPArcs(sip,sipArcPred):
                        #That is, each subquery contains values for the bound arguments
                        #that are passed through the sip arcs entering the node 
                        #corresponding to that literal
                        
                        #Create query out of body literal and apply sip-provided bindings
                        subquery = copy.deepcopy(bodyLiteral)
                        subquery.ground(projectedBindings)
                    if literalIsGround(subquery):
                        #subquery is ground, so there will only be boolean answers
                        #we return the conjunction of the answers for the current
                        #subquery
                        
                        answer = False
                        ns = None
                        
                        answers = first(
                                    itertools.dropwhile(
                                            lambda item:not item[0],
                                            SipStrategy(
                                                    subquery.toRDFTuple(),
                                                    sipCollection,
                                                    factGraph,
                                                    derivedPreds,
                                                    MakeImmutableDict(projectedBindings),
                                                    processedRules,
                                                    network = step is not None and \
                                                            step.parent.network or None,
                                                    debug = debug,
                                                    buildProof = buildProof,
                                                    memoizeMemory = memoizeMemory,
                                                    proofLevel = proofLevel)))
                        if answers:
                            answer,ns = answers
                        if not answer and not bodyLiteral.naf or \
                            (answer and bodyLiteral.naf):
                            #negative answer means the invokation of the rule fails
                            #either because we have a positive literal and there
                            #is no answer for the subgoal or the literal is 
                            #negative and there is an answer for the subgoal
                            raise RuleFailure("No solutions solving ground query %s"%subquery)
                        else:
                            if buildProof:
                                if not answer and bodyLiteral.naf:
                                    ns.naf = True
                                step.antecedents.append(ns)
                            #positive answer means we can continue processing the rule body
                            #either because we have a positive literal and answers
                            #for subgoal or a negative literal and no answers for the
                            #the goal
                            for rt,_step in invokeRule(
                                               [projectedBindings],
                                               bodyLiteralIterator,
                                               sip,
                                               otherargs,
                                               True,
                                               step,
                                               debug = debug):
                                yield rt,_step
                    else:
                        _answers = \
                                SipStrategy(subquery.toRDFTuple(),
                                            sipCollection,
                                            factGraph,
                                            derivedPreds,
                                            MakeImmutableDict(projectedBindings),
                                            processedRules,
                                            network = step is not None and \
                                                    step.parent.network or None,
                                            debug = debug,
                                            buildProof = buildProof,
                                            memoizeMemory = memoizeMemory,
                                            proofLevel = proofLevel)
                        #solve (non-ground) subgoal
                        def collectAnswers(_ans):
                            for ans,ns in _ans: 
                                if isinstance(ans,dict):
                                    try:
                                        map = mergeMappings1To2(ans,
                                                                projectedBindings,
                                                                makeImmutable=True)
                                        yield map
                                    except: pass
                        combinedAnswers = collectAnswers(_answers)
                        answers = lazyGeneratorPeek(combinedAnswers)
                        if not answers.successful and not bodyLiteral.naf or\
                          (bodyLiteral.naf and answers.successful):
                            raise RuleFailure("No solutions solving ground query %s"%subquery)
                        else:
                            #either we have a positive subgoal and answers
                            #or a negative subgoal and no answers
                            if buildProof:
                                if answers.successful:
                                    goals = set([g for a,g in answers])
                                    assert len(goals)==1
                                    step.antecedents.append(goals.pop())
                                else:
                                    newNs = NodeSet(
                                                bodyLiteral.toRDFTuple(),
                                                network=step.parent.network,
                                                identifier=BNode(),
                                                naf = True)
                                    step.antecedents.append(newNs)
                            for rt,_step in invokeRule(
                                               answers,
                                               bodyLiteralIterator,
                                               sip,
                                               otherargs,
                                               priorBooleanGoalSuccess,
                                               step,
                                               debug = debug,
                                               buildProof = buildProof):
                                yield rt,_step
            except StopIteration:
                #Finished processing rule
                if priorBooleanGoalSuccess:
                    yield projectedBindings and projectedBindings or True, step
                elif projectedBindings:
                    #Return the most recent (cumulative) answers and the given step
                    yield projectedBindings, step
                else:
                    raise RuleFailure("Finished processing rule unsuccessfully")
                
def refactorMapping(keyMapping,origMapping):
    """
    Takes a mapping from one mapping domain (D1)
    to another mapping domain (D2) as well as a mapping
    whose keys are in D1 and returns a new 
    """
    if keyMapping:
        refactoredMapping = {}
        for inKey,outKey in keyMapping.items(): 
            if inKey in origMapping:
                refactoredMapping[outKey]=origMapping[inKey]
        return refactoredMapping
    else:
        return origMapping
 
def prepMemiozedAns(ans):
    return isinstance(ans,dict) and MakeImmutableDict(ans) or ans
    

def SipStrategy(query,
                sipCollection,
                factGraph,
                derivedPreds,
                bindings={},
                processedRules = None,
                network = None,
                debug = False,
                buildProof = False,
                memoizeMemory = None,
                proofLevel = 1):
    """
    Accordingly, we define a sip-strategy for computing the answers to a query 
    expressed using a set of Datalog rules, and a set of sips, one for each 
    adornment of a rule head, as follows...
    
    Each evaluation uses memoization (via Python decorators) but also relies on well-formed 
    rewrites for using semi-naive bottom up method over large SPARQL data.
    
    """
    memoizeMemory = memoizeMemory and memoizeMemory or {}
    queryLiteral = BuildUnitermFromTuple(query)
    processedRules = processedRules and processedRules or set()
    if bindings:
        #There are bindings.  Apply them to the terms in the query
        queryLiteral.ground(bindings)
    
    if debug:
        print >> sys.stderr, "%sSolving"%('\t'*proofLevel), queryLiteral, bindings
    #Only consider ground triple pattern isomorphism with matching bindings
    goalRDFStatement = queryLiteral.toRDFTuple()

    if queryLiteral in memoizeMemory:
        if debug:
            print >> sys.stderr, "%sReturning previously calculated results for "%\
                        ('\t'*proofLevel), queryLiteral
        for answers in memoizeMemory[queryLiteral]:
            yield answers
    elif AlphaNode(goalRDFStatement).alphaNetworkHash(
                                      True,
                                      skolemTerms=bindings.values()) in\
        [AlphaNode(r.toRDFTuple()).alphaNetworkHash(True,
                                                    skolemTerms=bindings.values())
            for r in processedRules
                if AdornLiteral(goalRDFStatement).adornment == \
                   r.adornment]:
        if debug:
            print >> sys.stderr, "%s Goal already processed..."%\
                ('\t'*proofLevel)
    else: 
        isGround = literalIsGround(queryLiteral)
        if buildProof:
            ns=NodeSet(goalRDFStatement,network=network,identifier=BNode())    
        else:
            ns = None
        adornedProgram = factGraph.adornedProgram    
        queryPred = GetOp(queryLiteral)
        if sipCollection is None:
            rules = []
        else:
            #For every rule head matching the query, we invoke the rule, 
            #thus determining an adornment, and selecting a sip to follow            
            rules = sipCollection.headToRule.get(queryPred,set())
            if None in sipCollection.headToRule:
                #If there are second order rules, we add them
                #since they are a 'wildcard'
                rules.update(sipCollection.headToRule[None])

        #maintained list of rules that haven't been processed before and
        #match the query
        validRules = []
        
        #each subquery contains values for the bound arguments that are passed 
        #through the sip arcs entering the node corresponding to that literal. For
        #each subquery generated, there is a set of answers.
        answers = []
        
        variableMapping = {}
        
        #Some TBox queries can be 'joined' together into SPARQL queries against
        #'base' predicates via an RDF dataset
        #These atomic concept inclusion axioms can be evaluated together
        #using a disjunctive operator at the body of a horn clause
        #where each item is a query of the form uniPredicate(?X):
        #Or( uniPredicate1(?X1), uniPredicate2(?X), uniPredicate3(?X),..)
        #In this way massive, conjunctive joins can be 'mediated' 
        #between the stated facts and the top-down solver
        @parameterizedPredicate([i for i in derivedPreds])
        def IsAtomicInclusionAxiomRHS(rule,dPreds):
            """
            This is an atomic inclusion axiom with
            a variable (or bound) RHS:  uniPred(?ENTITY)
            """
            bodyList = list(iterCondition(rule.formula.body))
            body = first(bodyList)
            return GetOp(body) not in dPreds and \
                   len(bodyList) == 1 and \
                   body.op == RDF.type
        
        atomicInclusionAxioms = list(ifilter(IsAtomicInclusionAxiomRHS,rules))
        if atomicInclusionAxioms and len(atomicInclusionAxioms) > 1:
            if debug:
                print >> sys.stderr, "\tCombining atomic inclusion axioms: "
                pprint(atomicInclusionAxioms,sys.stderr)            
            if buildProof:
                factStep = InferenceStep(ns,source='some RDF graph')
                ns.steps.append(factStep)
                
            axioms = [rule.formula.body 
                      for rule in atomicInclusionAxioms]
     
            #attempt to exaustively apply any available substitutions
            #and determine if query if fully ground
            vars = [v for v in GetArgs(queryLiteral,
                                       secondOrder=True) 
                                             if isinstance(v,Variable)]
            openVars,axioms,_bindings  = \
                    normalizeBindingsAndQuery(vars,
                                              bindings,
                                              axioms)
            if openVars:
                mappings = {}
                #See if we need to do any variable mappings from the query literals
                #to the literals in the applicable rules
                query,rt = EDBQuery(axioms,
                                    factGraph,
                                    openVars,
                                    _bindings).evaluate(debug,
                                                        symmAtomicInclusion=True)
                if buildProof:
                    factStep.groundQuery = subquery
                for ans in rt:
                    if buildProof:
                        factStep.bindings.update(ans)
                    memoizeMemory.setdefault(queryLiteral,set()).add(
                                         (prepMemiozedAns(ans),ns))
                    yield ans, ns
            else:
                #All the relevant derivations have been explored and the result
                #is a ground query we can directly execute against the facts
                if buildProof:
                    factStep.bindings.update(bindings)
                query,rt = EDBQuery(axioms,
                                    factGraph,
                                    _bindings).evaluate(debug,
                                                        symmAtomicInclusion=True)                    
                if buildProof:
                    factStep.groundQuery = subquery
                memoizeMemory.setdefault(queryLiteral,set()).add(
                                         (prepMemiozedAns(rt),ns))                    
                yield rt,ns
            rules = ifilter(lambda i:not IsAtomicInclusionAxiomRHS(i),rules)                
        for rule in rules:
            #An exception is the special predicate ph; it is treated as a base 
            #predicate and the tuples in it are those supplied for qb by unification.
            headBindings = getBindingsFromLiteral(goalRDFStatement,rule.formula.head)
            comboBindings = dict([(k,v) for k,v in itertools.chain(
                                                      bindings.items(),
                                                      headBindings.items())])
            varMap = rule.formula.head.getVarMapping(queryLiteral)
            if headBindings and\
                [term for term in rule.formula.head.getDistinguishedVariables(True)
                        if varMap.get(term,term) not in headBindings]:
                continue
            subQueryAnswers = []
            dontStop = True
            projectedBindings = comboBindings.copy()
            if debug:
                print >> sys.stderr, "%sProcessing rule"%\
                ('\t'*proofLevel), rule.formula
                
            try:
                #Invoke the rule
                if buildProof:
                    step = InferenceStep(ns,rule.formula)
                else:
                    step = None
                for rt,step in\
                  invokeRule([headBindings],
                              iter(iterCondition(rule.formula.body)),
                              rule.sip,
                              (proofLevel + 1,
                               memoizeMemory,
                               sipCollection, 
                               factGraph, 
                               derivedPreds,
                               processedRules.union([
                                 AdornLiteral(query)])),
                              step=step,
                              debug = debug):
                    if rt:
                        if isinstance(rt,dict):
                            #We received a mapping and must rewrite it via
                            #correlation between the variables in the rule head
                            #and the variables in the original query (after applying 
                            #bindings) 
                            varMap = rule.formula.head.getVarMapping(queryLiteral)
                            if varMap:
                                rt = MakeImmutableDict(refactorMapping(varMap,rt))
                            if buildProof:
                                step.bindings = rt
                        else:
                            if buildProof:
                                step.bindings = headBindings
                        validRules.append(rule)
                        if buildProof:
                            ns.steps.append(step)
                        if isGround:
                            yield True,ns
                        else:
                            memoizeMemory.setdefault(queryLiteral,set()).add(
                                                     (prepMemiozedAns(rt),
                                                      ns))        
                            yield rt, ns 
                
            except RuleFailure, e:
                #Clean up failed antecedents
                if buildProof:
                    if ns in step.antecedents:
                        step.antecedents.remove(ns)
        if not validRules:
            #No rules matching, query factGraph for answers
            successful = False
            if buildProof:
                factStep = InferenceStep(ns,source='some RDF graph')
                ns.steps.append(factStep)
            if not isGround:
                subquery,rt = EDBQuery([queryLiteral],
                                        factGraph,
                                        [v for v in GetArgs(queryLiteral,
                                                            secondOrder=True)
                                                                 if isinstance(v,Variable)],

                                        bindings).evaluate(debug)                    
                if buildProof:
                    factStep.groundQuery = subquery
                for ans in rt:
                    successful =True
                    if buildProof:
                        factStep.bindings.update(ans)
                    memoizeMemory.setdefault(queryLiteral,set()).add(
                                             (prepMemiozedAns(ans),
                                              ns))
                    yield ans, ns
                if not successful and queryPred not in derivedPreds:
                    #Open query didn't return any results and the predicate
                    #is ostensibly marked as derived predicate, so we have failed
                    memoizeMemory.setdefault(queryLiteral,set()).add((False,ns))
                    yield False,ns
            else:
                #All the relevant derivations have been explored and the result
                #is a ground query we can directly execute against the facts
                if buildProof:
                    factStep.bindings.update(bindings)

                subquery,rt = EDBQuery([queryLiteral],
                                        factGraph,
                                        bindings).evaluate(debug)                    
                if buildProof:
                    factStep.groundQuery = subquery
                memoizeMemory.setdefault(queryLiteral,set()).add(
                                             (prepMemiozedAns(rt),
                                              ns))                    
                yield rt,ns

def test():
     import doctest
     doctest.testmod()    
if __name__ == '__main__':
    test()


#!/usr/bin/env python
# encoding: utf-8
"""
Helper Functions for reducing DL axioms into a normal forms
"""
from cStringIO import StringIO
from rdflib.Graph import Graph
from rdflib.util import first
from rdflib import URIRef, RDF, RDFS, Namespace, Variable, Literal, URIRef, BNode
from rdflib.syntax.NamespaceManager import NamespaceManager
from FuXi.Rete import ReteNetwork
from FuXi.Rete.RuleStore import N3RuleStore
from FuXi.Rete.Util import generateTokenSet
from FuXi.Syntax.InfixOWL import *
from FuXi.DLP import SKOLEMIZED_CLASS_NS
#from FuXi.Rete.Magic import MagicSetTransformation
import sys, unittest

class NominalRangeTransformer(object):        
    NOMINAL_QUERY=\
    """SELECT ?RESTRICTION ?INTERMEDIATE_CLASS ?NOMINAL ?PROP
       { ?RESTRICTION owl:onProperty ?PROP;
                      owl:someValuesFrom ?INTERMEDIATE_CLASS .  
         ?INTERMEDIATE_CLASS owl:oneOf ?NOMINAL .  } """    
    
    def transform(self,graph):
        """
        Transforms a 'pure' nominal range into a disjunction of value restrictions
        """
        Individual.factoryGraph = graph
        for restriction,intermediateCl,nominal,prop in graph.query(
                                 self.NOMINAL_QUERY,
                                 initNs={u'owl':OWL_NS}):
            nominalCollection=Collection(graph,nominal)
            #purge restriction
            restr=Class(restriction)
            parentSets = [i for i in restr.subClassOf]
            restr.clearOutDegree()
            newConjunct=BooleanClass(restriction,
                                     OWL_NS.unionOf,
                                     [Property(prop)|value|val 
                                                 for val in nominalCollection],
                                     graph)
            newConjunct.subClassOf = parentSets
            
            #purge nominalization placeholder
            iClass = BooleanClass(intermediateCl)
            iClass.clear()
            iClass.delete()

class UniversalNominalRangeTransformer(object):        
    NOMINAL_QUERY=\
    """SELECT ?RESTRICTION ?INTERMEDIATE_CLASS ?NOMINAL ?PROP ?PARTITION
       { ?RESTRICTION owl:onProperty ?PROP;
                      owl:allValuesFrom ?INTERMEDIATE_CLASS .  
         ?INTERMEDIATE_CLASS owl:oneOf ?NOMINAL .
         ?PROP rdfs:range [ owl:oneOf ?PARTITION ] .
       } """    
    
    def transform(self,graph):
        """
        Transforms a universal restriction on a 'pure' nominal range into a 
        conjunction of value restriction (using set theory and demorgan's laws)
        """
        Individual.factoryGraph = graph
        for restriction,intermediateCl,nominal,prop, partition in graph.query(
                                 self.NOMINAL_QUERY,
                                 initNs={u'owl':OWL_NS,
                                         u'rdfs':RDFS.RDFSNS}):
            exceptions = EnumeratedClass()
            partition = Collection(graph,partition)
            nominalCollection=Collection(graph,nominal)
            for i in partition:
                if i not in nominalCollection:
                    exceptions._rdfList.append(i)
                    #exceptions+=i 
            exists=Class(complementOf=(Property(prop)|some|exceptions))
            for s,p,o in graph.triples((None,None,restriction)):
                graph.add((s,p,exists.identifier))
            Individual(restriction).delete()

            #purge nominalization placeholder
            iClass = BooleanClass(intermediateCl)
            iClass.clear()
            iClass.delete()
            
class GeneralUniversalTransformer(object):
    def transform(self,graph):
        """
        Transforms a universal restriction to a negated existential restriction
        """
        Individual.factoryGraph = graph
        for restr,p,o in graph.triples((None,OWL_NS.allValuesFrom,None)):
            graph.remove((restr,p,o))
            innerCompl = Class(complementOf=o)
            graph.add((restr,OWL_NS.someValuesFrom,innerCompl.identifier))
            outerCompl = Class()
            for _s,_p,_o in graph.triples((None,None,restr)):
                graph.add((_s,_p,outerCompl.identifier))
                graph.remove((_s,_p,_o))
            outerCompl.complementOf=restr

class DoubleNegativeTransformer(object):
    UNIVERSAL_QUERY=\
    """SELECT ?COMPL1 ?COMPL2 ?COMPL3
       { ?COMPL1 owl:complementOf ?COMPL2 .
         ?COMPL2 owl:complementOf ?COMPL3
         FILTER( isBlank(?COMPL1) && isBlank(?COMPL2) )
       } """
    
    def transform(self,graph):
        Individual.factoryGraph = graph
        for compl1,compl2,compl3 in graph.query(
                                 self.UNIVERSAL_QUERY,
                                 initNs={u'owl':OWL_NS,
                                         u'rdfs':RDFS.RDFSNS}):
            Individual(compl1).replace(compl3)
            Individual(compl2).delete()

            
class DemorganTransformer(object):        
    def transform(self,graph):
        """
        Uses demorgan's laws to reduce negated disjunctions to a conjunction of
        negated formulae
        """
        Individual.factoryGraph = graph
        for disjunctId in graph.subjects(predicate=OWL_NS.unionOf):
            if (None,OWL_NS.complementOf,disjunctId) in graph and \
               isinstance(disjunctId,BNode):
                #not (     A1 or      A2  or .. or      An ) 
                #                 = 
                #    ( not A1 and not A2 and .. and not An )
                disjunct = BooleanClass(disjunctId,operator=OWL_NS.unionOf)
                items = list(disjunct)
                newConjunct = BooleanClass(members=[~Class(item) for item in items])
                for negation in graph.subjects(predicate=OWL_NS.complementOf,
                                               object=disjunctId):
                    Class(negation).replace(newConjunct)
                    if not isinstance(negation,BNode):
                        newConjunct.identifier = negation        
                disjunct.clear()
                disjunct.delete()
            elif ((disjunctId,OWL_NS.unionOf,None) in graph) and not \
                   [item for item in BooleanClass(disjunctId,
                                                  operator=OWL_NS.unionOf)
                        if not Class(item).complementOf]:
                #( not A1 or  not A2  or .. or  not An ) 
                #                 = 
                #not ( A1 and A2 and .. and An )                         
                disjunct = BooleanClass(disjunctId,operator=OWL_NS.unionOf)
                items = [Class(item).complementOf for item in disjunct]
                for negation in disjunct:
                    Class(negation).delete()
                negatedConjunct = ~ BooleanClass(members=items)
                disjunct.clear()
                disjunct.replace(negatedConjunct)

class ConjunctionFlattener(object):        
    def transform(self,graph):
        """
        Flattens conjunctions
        ( A1 and ( B1 and B2 ) and A2 ) 
                         = 
        ( A1 and B1 and B2 and A2 )
        
        """
        Individual.factoryGraph = graph
        for conjunctId in graph.subjects(predicate=OWL_NS.intersectionOf):
            conjunct = BooleanClass(conjunctId)
            nestedConjuncts = [BooleanClass(i) for i in conjunct
                                    if (i,OWL_NS.intersectionOf,None) in graph]
            if nestedConjuncts:
                def collapseConjunctTerms(left,right):
                    list(left)+list(right)
                if len(nestedConjuncts) == 1:
                    newTopLevelItems = list(nestedConjuncts[0])
                else:
                    newTopLevelItems = reduce(collapseConjunctTerms,nestedConjuncts)
                for nc in nestedConjuncts:
                    nc.clear()
                    del conjunct[conjunct.index(nc.identifier)]
                    nc.delete()
                for newItem in newTopLevelItems:
                    conjunct.append(newItem)
            
def NormalFormReduction(ontGraph):
    UniversalNominalRangeTransformer().transform(ontGraph)
    GeneralUniversalTransformer().transform(ontGraph)
    DoubleNegativeTransformer().transform(ontGraph)    
    NominalRangeTransformer().transform(ontGraph)
    DemorganTransformer().transform(ontGraph)
    DoubleNegativeTransformer().transform(ontGraph)    
    ConjunctionFlattener().transform(ontGraph)

EX_NS = Namespace('http://example.com/')
EX    = ClassNamespaceFactory(EX_NS)

class ReductionTestA(unittest.TestCase):
    def setUp(self):
        self.ontGraph = Graph()
        self.ontGraph.bind('ex', EX_NS)
        self.ontGraph.bind('owl', OWL_NS)
        Individual.factoryGraph = self.ontGraph
        partition = EnumeratedClass(EX_NS.part,
                                    members=[EX_NS.individual1,
                                             EX_NS.individual2,
                                             EX_NS.individual3])
        subPartition = EnumeratedClass(EX_NS.partition,members=[EX_NS.individual1])
        partitionProp = Property(EX_NS.propFoo,
                                 range=partition)
        self.foo = EX.foo
        self.foo.subClassOf = [partitionProp|only|subPartition] 
        
    def testUnivInversion(self):
        UniversalNominalRangeTransformer().transform(self.ontGraph)
        self.failUnlessEqual(len(list(self.foo.subClassOf)),
                             1,
                             "There should still be one subsumed restriction")
        subC = CastClass(first(self.foo.subClassOf))
        self.failUnless(not isinstance(subC,Restriction),
                        "subclass of a restriction")
        self.failUnless(subC.complementOf is not None,"Should be a complement!")
        innerC = CastClass(subC.complementOf)
        self.failUnless(isinstance(innerC,Restriction),
                        "complement of a restriction, not %r"%innerC)
        self.failUnlessEqual(innerC.onProperty,
                             EX_NS.propFoo,
                             "restriction on propFoo")
        self.failUnless(innerC.someValuesFrom,"converted to an existential restriction not %r"%innerC)
        invertedC = CastClass(innerC.someValuesFrom)
        self.failUnless(isinstance(invertedC,EnumeratedClass),
                        "existencial restriction on enumerated class")
        self.assertEqual(len(invertedC),
                         2,
                        "existencial restriction on enumerated class of length 2")
        self.assertEqual(repr(invertedC),
                         "{ ex:individual2 ex:individual3 }",
                         "The negated partition should exclude individual1")
        NominalRangeTransformer().transform(self.ontGraph)
        DemorganTransformer().transform(self.ontGraph)
        
        subC = CastClass(first(self.foo.subClassOf))
        self.assertEqual(repr(subC),
                        "( ( not ( ex:propFoo value ex:individual2 ) ) and ( not ( ex:propFoo value ex:individual3 ) ) )")

class ReductionTestB(unittest.TestCase):
    def setUp(self):
        self.ontGraph = Graph()
        self.ontGraph.bind('ex', EX_NS)
        self.ontGraph.bind('owl', OWL_NS)
        Individual.factoryGraph = self.ontGraph
        disjunct = (~ EX.alpha) | (~ EX.omega)
        self.foo = EX.foo
        disjunct+=self.foo 
        
    def testHiddenDemorgan(self):
        NormalFormReduction(self.ontGraph)
        self.failUnless(first(self.foo.subClassOf).complementOf,
                        "should be the negation of a boolean class")
        innerC = CastClass(first(self.foo.subClassOf).complementOf)
        self.failUnless(isinstance(innerC,BooleanClass) and \
                        innerC._operator == OWL_NS.intersectionOf,
                        "should be the negation of a conjunct")        
        self.assertEqual(repr(innerC),"( ex:alpha and ex:omega )")

class FlatteningTest(unittest.TestCase):
    def setUp(self):
        self.ontGraph = Graph()
        self.ontGraph.bind('ex', EX_NS)
        self.ontGraph.bind('owl', OWL_NS)
        Individual.factoryGraph = self.ontGraph
        nestedConjunct = EX.omega & EX.gamma
        self.topLevelConjunct = EX.alpha & nestedConjunct
        
    def testFlatenning(self):
        self.assertEquals(repr(self.topLevelConjunct),
                          "ex:alpha that ( ex:omega and ex:gamma )")
        ConjunctionFlattener().transform(self.ontGraph)
        self.assertEquals(repr(self.topLevelConjunct),
                          "( ex:alpha and ex:omega and ex:gamma )")

class UniversalComplementXFormTest(unittest.TestCase):
    def setUp(self):
        self.ontGraph = Graph()
        self.ontGraph.bind('ex', EX_NS)
        self.ontGraph.bind('owl', OWL_NS)
        Individual.factoryGraph = self.ontGraph

    def testUniversalInversion(self):
        testClass1 = EX.omega & (Property(EX_NS.someProp)|only|~EX.gamma)
        testClass1.identifier = EX_NS.Foo
        self.assertEquals(repr(testClass1),
                          "ex:omega that ( ex:someProp only ( not ex:gamma ) )")
        NormalFormReduction(self.ontGraph)
        self.assertEquals(repr(testClass1),
                          "ex:omega that ( not ( ex:someProp some ex:gamma ) )")
    
if __name__ == '__main__':
    unittest.main()


"""
Utility functions for a Boost Graph Library (BGL) DiGraph via the BGL Python Bindings
"""
import itertools, pickle
from FuXi.Rete.AlphaNode import AlphaNode        
from rdflib.Graph import Graph
from rdflib.syntax.NamespaceManager import NamespaceManager
from rdflib import BNode, Namespace, Collection, Variable, URIRef
from rdflib.util import first
try:    
    import boost.graph as bgl
    bglGraph = bgl.Digraph()
except:
    try:
        from pydot import Node,Edge,Dot
    except:
        import warnings
        warnings.warn("Missing pydot library",ImportWarning)        

LOG = Namespace("http://www.w3.org/2000/10/swap/log#")

def xcombine(*seqin):
    '''
    http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/302478
    returns a generator which returns combinations of argument sequences
    for example xcombine((1,2),(3,4)) returns a generator; calling the next()
    method on the generator will return [1,3], [1,4], [2,3], [2,4] and
    StopIteration exception.  This will not create the whole list of 
    combinations in memory at once.
    '''
    def rloop(seqin,comb):
        '''recursive looping function'''
        if seqin:                   # any more sequences to process?
            for item in seqin[0]:
                newcomb=comb+[item]     # add next item to current combination
                # call rloop w/ remaining seqs, newcomb
                for item in rloop(seqin[1:],newcomb):   
                    yield item          # seqs and newcomb
        else:                           # processing last sequence
            yield comb                  # comb finished, add to list
    return rloop(seqin,[])

def permu(xs):
    """
    http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/496819
    "A recursive function to get permutation of a list"
    
    >>> print list(permu([1,2,3]))
    [[1, 2, 3], [1, 3, 2], [2, 1, 3], [2, 3, 1], [3, 1, 2], [3, 2, 1]]
     
    """
    if len(xs) <= 1:
        yield xs
    else:
        for i in range(len(xs)):
            for p in permu(xs[:i] + xs[i + 1:]):
                yield [xs[i]] + p

def CollapseDictionary(mapping):
    """
    Takes a dictionary mapping prefixes to URIs
    and removes prefix mappings that begin with _ and
    there is already a map to their value
    
    >>> a = {'ex': URIRef('http://example.com/')}
    >>> a['_1'] =  a['ex']
    >>> len(a)
    2
    >>> a.values()
    [rdflib.URIRef('http://example.com/'), rdflib.URIRef('http://example.com/')]
    >>> CollapseDictionary(a)
    {'ex': rdflib.URIRef('http://example.com/')}
    >>> a
    {'ex': rdflib.URIRef('http://example.com/'), '_1': rdflib.URIRef('http://example.com/')}
    """
    def originalPrefixes(item):
        return item.find('_')+1==1
    revDict = {}
    for k,v in mapping.items():
        revDict.setdefault(v,set()).add(k)
    prefixes2Collapse = []
    for k,v in revDict.items():
        origPrefixes=[]
        dupePrefixes=[]
        #group prefixes for a single URI by whether or not
        #they have a _ prefix
        for rt,items in itertools.groupby(v,originalPrefixes):
            if rt:
                dupePrefixes.extend(items)
            else:
                origPrefixes.extend(items)
        if origPrefixes and len(v) > 1 and len(dupePrefixes):
            #There are allocated prefixes for URIs that were originally
            #given a prefix 
            assert len(origPrefixes)==1
            prefixes2Collapse.extend(dupePrefixes)
    return dict([(k,v) for k,v in mapping.items() if k not in prefixes2Collapse])

class selective_memoize(object):
    """Decorator that caches a function's return value each time it is called.
    If called later with the same arguments, the cached value is returned, and
    not re-evaluated. Slow for mutable types.
    The arguments used for the cache are given to the decorator
    
    >>> @selective_memoize([0,1])
    ... def addition(l,r,other): 
    ...     print "calculating.."
    ...     return l+r
    >>> addition(1,2,3)
    calculating..
    3
    >>> addition(1,2,4)
    3
    >>> @selective_memoize()
    ... def addition(l,r,other): 
    ...     print "calculating.."
    ...     return l+r    
    >>> addition(1,2,3)
    calculating..
    3
    >>> addition(1,2,4)
    calculating..
    3
    >>> @selective_memoize([0,1],'baz')
    ... def addition(l,r,baz=False, bar=False): 
    ...     print "calculating.."
    ...     return l+r    
    >>> addition(1,2,baz=True)
    calculating..
    3
    >>> addition(1,2,baz=True,bar = True)
    3
    """
    # Ideas from MemoizeMutable class of Recipe 52201 by Paul Moore and
    # from memoized decorator of http://wiki.python.org/moin/PythonDecoratorLibrary
    def __init__(self, cacheableArgPos = [], cacheableArgKey = []):
        self.cacheableArgPos = cacheableArgPos
        self.cacheableArgKey = cacheableArgKey
        self._cache = {}
        
    def __call__(self, func):
        def innerHandler(*args, **kwds):
            if self.cacheableArgPos:
                chosenKeys = []
                for idx,arg in enumerate(args):
                    if idx in self.cacheableArgPos:
                        chosenKeys.append(arg)
                key = tuple(chosenKeys)
            else:
                key = args
            if kwds:
                if self.cacheableArgKey:
                    items = [(k,v) for k,v in kwds.items() 
                                if k in self.cacheableArgKey]
                else:
                    items = []
                items.sort()
                key = key + tuple(items)
            try:
                if key in self._cache:
                    return self._cache[key]
                self._cache[key] = result = func(*args, **kwds)
                return result
            except TypeError:
                try:
                    dump = pickle.dumps(key)
                except pickle.PicklingError:
                    return func(*args, **kwds)
                else:
                    if dump in self._cache:
                        return self._cache[dump]
                    self._cache[dump] = result = func(*args, **kwds)
                    return result
        return innerHandler

class InformedLazyGenerator(object):
    def __init__(self, generator, successful):
        self.generator  = generator
        self.successful = successful
        
    def __iter__(self):
        for item in self.generator:
            yield item        
    
def lazyGeneratorPeek(iterable,firstN=1):
    """
    Lazily peeks into an iterable and returns None if it has less than N items 
    or returns another generator over *all* content if it isn't
    
    >>> a=(i for i in [1,2,3])
    >>> first(a)
    1
    >>> list(a)
    [2, 3]
    >>> a=(i for i in [1,2,3])
    >>> result = lazyGeneratorPeek(a)
    >>> result  # doctest:+ELLIPSIS
    <generator object at ...>
    >>> result = list(result)
    >>> result
    [1, 2, 3]
    >>> lazyGeneratorPeek((i for i in []))
    >>> lazyGeneratorPeek(result,4)
    >>> lazyGeneratorPeek(result,3) # doctest:+ELLIPSIS
    <generator object at ...>
    """
    cnt = firstN
    header = []
    for item in iterable:
        cnt -= 1
        header.append(item)
        if not cnt:
            #Stop after consuming first N items
            break
    if not cnt:
        #There at least N items
        return InformedLazyGenerator((i for i in itertools.chain(header,iterable)),True)
    else:
        return InformedLazyGenerator((i for i in header),False)

def generateTokenSet(graph,debugTriples=[],skipImplies=True):
    """
    Takes an rdflib graph and generates a corresponding Set of ReteTokens
    Note implication statements are excluded from the realm of facts by default
    """
    from FuXi.Rete import ReteToken
    rt = set()    
    def normalizeGraphTerms(term):
        if isinstance(term,Collection.Collection):
            return term.uri
        else:
            return term
    for s,p,o in graph:
        
        if not skipImplies or p != LOG.implies:
            #print s,p,o             
            debug = debugTriples and (s,p,o) in debugTriples
            rt.add(ReteToken((normalizeGraphTerms(s),
                              normalizeGraphTerms(p),
                              normalizeGraphTerms(o)),debug))
    return rt

def generateBGLNode(dot,node,namespace_manager,identifier):
    from FuXi.Rete import ReteNetwork,BetaNode,BuiltInAlphaNode,AlphaNode
    from BetaNode import LEFT_MEMORY, RIGHT_MEMORY, LEFT_UNLINKING
    vertex = Node(identifier)

    shape = 'circle'
    root        = False
    if isinstance(node,ReteNetwork):     
        root        = True
        peripheries = '3'
    elif isinstance(node,BetaNode) and not node.consequent:     
        peripheries = '1'
        if node.fedByBuiltin:
            label = "Built-in pass-thru\\n"
        elif node.aPassThru:
            label = "Pass-thru Beta node\\n"
        elif node.commonVariables:
            label = "Beta node\\n(%s)"%(','.join(["?%s"%i for i in node.commonVariables]))
        else:
            label = "Beta node"
        if not node.fedByBuiltin:
            leftLen = node.memories[LEFT_MEMORY] and len(node.memories[LEFT_MEMORY]) or 0
            rightLen = len(node.memories[RIGHT_MEMORY])
            label+='\\n %s in left, %s in right memories'%(leftLen,rightLen)

    elif isinstance(node,BetaNode) and node.consequent:     
        #rootMap[vertex] = 'true'
        peripheries = '2'
        stmts = []
        for s,p,o in node.consequent:
            stmts.append(' '.join([str(namespace_manager.normalizeUri(s)),
              str(namespace_manager.normalizeUri(p)),
              str(namespace_manager.normalizeUri(o))]))
              
        rhsVertex = Node(BNode(),
                         label='"'+'\\n'.join(stmts)+'"',
                         shape='plaintext') 
        edge = Edge(vertex,rhsVertex)
        #edge.color = 'red'
        dot.add_edge(edge)              
        dot.add_node(rhsVertex)      
        if node.commonVariables:
            inst=node.network.instanciations.get(node,0)
            label = str("Terminal node\\n(%s)\\n%d instanciations"%(','.join(["?%s"%i for i in node.commonVariables]),inst))
        else:
            label = "Terminal node"
        leftLen = node.memories[LEFT_MEMORY] and len(node.memories[LEFT_MEMORY]) or 0
        rightLen = len(node.memories[RIGHT_MEMORY])
        label+='\\n %s in left, %s in right memories'%(leftLen,rightLen)      
        inst = node.network.instanciations[node]
        if inst:
            label += "\\n%s instanciations"%inst      
        
    elif isinstance(node,BuiltInAlphaNode):
        peripheries = '1'
        shape = 'plaintext'
        #label = '..Builtin Source..'
        label = repr(node.n3builtin)
        canonicalFunc = namespace_manager.normalizeUri(node.n3builtin.uri)
        canonicalArg1 = namespace_manager.normalizeUri(node.n3builtin.argument)
        canonicalArg2 = namespace_manager.normalizeUri(node.n3builtin.result)
        label = '%s(%s,%s)'%(canonicalFunc,canonicalArg1,canonicalArg2)
        
    elif isinstance(node,AlphaNode):
        peripheries = '1'
        shape = 'plaintext'
#        widthMap[vertex] = '50em'
        label = ' '.join([isinstance(i,BNode) and i.n3() or str(namespace_manager.normalizeUri(i)) 
                           for i in node.triplePattern])    

    vertex.set_shape(shape)
    vertex.set_label('"%s"'%label)
    vertex.set_peripheries(peripheries)
    if root:
        vertex.set_root('true')
    return vertex

def renderNetwork(network,nsMap = {}):
    """
    Takes an instance of a compiled ReteNetwork and a namespace mapping (for constructing QNames
    for rule pattern terms) and returns a BGL Digraph instance representing the Rete network
    #(from which GraphViz diagrams can be generated)
    """
    from FuXi.Rete import BuiltInAlphaNode
    from BetaNode import LEFT_MEMORY, RIGHT_MEMORY, LEFT_UNLINKING
    dot=Dot(graph_type='digraph')
    namespace_manager = NamespaceManager(Graph())
    for prefix,uri in nsMap.items():
        namespace_manager.bind(prefix, uri, override=False)

    visitedNodes = {}
    edges = []
    idx = 0
    for node in network.nodes.values():
        if not node in visitedNodes:
            idx += 1
            visitedNodes[node] = generateBGLNode(dot,node,namespace_manager,str(idx))
            dot.add_node(visitedNodes[node])
    nodeIdxs = {}                        
    for node in network.nodes.values():
        for mem in node.descendentMemory:
            if not mem:
                continue
            bNode = mem.successor
        for bNode in node.descendentBetaNodes:
            for idx,otherNode in enumerate([bNode.leftNode,bNode.rightNode]):
                if node == otherNode and (node,otherNode) not in edges:
                    for i in [node,bNode]:
                        if i not in visitedNodes:
                            idx += 1
                            nodeIdxs[i] = idx 
                            visitedNodes[i] = generateBGLNode(dot,i,namespace_manager,str(idx))
                            dot.add_node(visitedNodes[i])
                    edge = Edge(visitedNodes[node],
                                visitedNodes[bNode],
                                label=idx==0 and 'left' or 'right')
                    dot.add_edge(edge)                                        
                    edges.append((node,bNode))
                    
    return dot

def test():
    import doctest
    doctest.testmod()

if __name__ == '__main__':
    test()    


#!/usr/bin/env python
from pprint import pprint
from FuXi.Rete.Proof import GenerateProof
from FuXi.Rete import ReteNetwork
from FuXi.Rete.AlphaNode import SUBJECT,PREDICATE,OBJECT,VARIABLE
from FuXi.Rete.BetaNode import PartialInstanciation, LEFT_MEMORY, RIGHT_MEMORY
from FuXi.Rete.RuleStore import N3RuleStore, SetupRuleStore
from FuXi.Rete.Util import renderNetwork,generateTokenSet, xcombine
from FuXi.DLP.DLNormalization import NormalFormReduction
from FuXi.LP.BackwardFixpointProcedure import BackwardFixpointProcedure
from FuXi.LP import IdentifyHybridPredicates
from FuXi.SPARQL.BackwardChainingStore import * 
from FuXi.DLP import MapDLPtoNetwork, DisjunctiveNormalForm
from FuXi.Horn import *
from FuXi.Horn.HornRules import HornFromN3, Ruleset
from FuXi.Syntax.InfixOWL import *
from FuXi.Rete.TopDown import *
from FuXi.Rete.Proof import ProofBuilder, PML, GMP_NS
from FuXi.Rete.Magic import *
from FuXi.Rete.SidewaysInformationPassing import *
from rdflib.sparql.bison.Query import Prolog
from rdflib.Namespace import Namespace
from rdflib import plugin,RDF,RDFS,URIRef,URIRef,Literal,Variable
from rdflib.store import Store
from cStringIO import StringIO
from rdflib.Graph import Graph,ReadOnlyGraphAggregate,ConjunctiveGraph
from rdflib.syntax.NamespaceManager import NamespaceManager
from rdflib.sparql.parser import parse as ParseSPARQL
from rdflib.sparql.Algebra import ReduceGraphPattern
import unittest, time, warnings,sys

TEMPLATES = Namespace('http://code.google.com/p/fuxi/wiki/BuiltinSPARQLTemplates#')

def main():
    from optparse import OptionParser
    op = OptionParser('usage: %prog [options] factFile1 factFile2 ... factFileN')
    op.add_option('--why', 
                  default=None,
      help = 'Specifies the goals to solve for using the non-niave methods'+
              'see --method')    
    op.add_option('--closure', 
                  action='store_true',
                  default=False,
      help = 'Whether or not to serialize the inferred triples'+ 
             ' along with the original triples.  Otherwise '+
              '(the default behavior), serialize only the inferred triples')
    op.add_option('--output', 
                  default='n3',
                  metavar='RDF_FORMAT',
                  choices = ['xml', 
                             'TriX', 
                             'n3', 
                             'pml',
                             'proof-graph',
                             'nt',
                             'rif',
                             'rif-xml',
                             'conflict',
                             'man-owl'],
      help = "Serialize the inferred triples and/or original RDF triples to STDOUT "+
             "using the specified RDF syntax ('xml','pretty-xml','nt','turtle', "+
             "or 'n3') or to print a summary of the conflict set (from the RETE "+
             "network) if the value of this option is 'conflict'.  If the the "+
             " value is 'rif' or 'rif-xml', Then the rules used for inference "+
             "will be serialized as RIF.  If the value is 'pml' and --why is used, "+
             " then the PML RDF statements are serialized.  If output is "+
             "'proof-graph then a graphviz .dot file of the proof graph is printed. "+
             "Finally if the value is 'man-owl', then the RDF facts are assumed "+
             "to be OWL/RDF and serialized via Manchester OWL syntax. The default is %default")
    op.add_option('--class',
                  dest='classes',
                  action='append',
                  default=[],
                  metavar='QNAME', 
      help = 'Used with --output=man-owl to determine which '+
             'classes within the entire OWL/RDF are targetted for serialization'+
             '.  Can be used more than once')
    op.add_option('--hybrid',
                  action='store_true',
                  default=False,
      help = 'Used with with --method=bfp to determine whether or not to '+
             'peek into the fact graph to identify predicates that are both '+
             'derived and base.  This is expensive for large fact graphs'+
             'and is explicitely not used against SPARQL endpoints')
    op.add_option('--property',
                  action='append',
                  dest='properties',
                  default=[],
                  metavar='QNAME', 
      help = 'Used with --output=man-owl or --extract to determine which '+
             'properties are serialized / extracted.  Can be used more than once')
    op.add_option('--normalize', 
                  action='store_true',
                  default=False,
      help = "Used with --output=man-owl to attempt to determine if the ontology is 'normalized' [Rector, A. 2003]"+
      "The default is %default")
    op.add_option('--ddlGraph', 
                default=False,
      help = "The location of a N3 Data Description document describing the IDB predicates")      
    op.add_option('--input-format', 
                  default='xml',
                  dest='inputFormat',
                  metavar='RDF_FORMAT',
                  choices = ['xml', 'trix', 'n3', 'nt', 'rdfa'],
      help = "The format of the RDF document(s) which serve as the initial facts "+
             " for the RETE network. One of 'xml','n3','trix', 'nt', "+
             "or 'rdfa'.  The default is %default")
    op.add_option('--safety', 
                  default='none',
                  metavar='RULE_SAFETY',
                  choices = ['loose', 'strict','none'],
      help = "Determines how to handle RIF Core safety.  A value of 'loose' "+
             " means that unsafe rules will be ignored.  A value of 'strict' "+
             " will cause a syntax exception upon any unsafe rule.  A value of "+
             "'none' (the default) does nothing")    
    op.add_option('--pDSemantics', 
                  action='store_true',
                  default=False,
      help = 'Used with --dlp to add pD semantics ruleset for semantics not covered '+
      'by DLP but can be expressed in definite Datalog Logic Programming'+
      ' The default is %default')
    op.add_option('--stdin', 
                  action='store_true',
                  default=False,
      help = 'Parse STDIN as an RDF graph to contribute to the initial facts. The default is %default ')
    op.add_option('--ns', 
                  action='append',
                  default=[],
                  metavar="PREFIX=URI",
      help = 'Register a namespace binding (QName prefix to a base URI).  This '+
             'can be used more than once')
    op.add_option('--rules', 
                  default=[],
                  action='append',
                  metavar='PATH_OR_URI',
      help = 'The Notation 3 documents to use as rulesets for the RETE network'+
      '.  Can be specified more than once')
    op.add_option('-d', '--debug', action='store_true',default=False,
      help = 'Include debugging output')
    op.add_option('--strictness',
                  default='defaultBase',
                  metavar='DDL_STRICTNESS',
                  choices = ['loose', 
                             'defaultBase', 
                             'defaultDerived', 
                             'harsh'],
      help = 'Used with --why to specify whether to: *not* check if predicates are '+
      ' both derived and base (loose), if they are, mark as derived (defaultDerived) '+
      'or as base (defaultBase) predicates, else raise an exception (harsh)')    
    op.add_option('--method',
                  default='naive',
                  metavar='reasoning algorithm',
                  choices = ['gms','sld','bfp','naive'],
      help = 'Used with --why to specify how to evaluate answers for query.  '+
      'Either both (default), top-down alone, or bottom-up alone')        
    op.add_option('--firstAnswer',
                  default=False,
                  action='store_true',
      help = 'Used with --why to determine whether to fetch all answers or just '+
      'the first')            
    op.add_option('--edb',
                  default=[],
                  action='append',
                  metavar='EXTENSIONAL_DB_PREDICATE_QNAME',                  
      help = 'Used with --why/--strictness=defaultDerived to specify which clashing '+
      'predicate will be designated as a base predicate')
    op.add_option('--idb',
                  default=[],
                  action='append',
                  metavar='INTENSIONAL_DB_PREDICATE_QNAME',                  
      help = 'Used with --why/--strictness=defaultBase to specify which clashing '+
      'predicate will be designated as a derived predicate')                    
    op.add_option('--noMagic',
                  default=[],
                  action='append',
                  metavar='DB_PREDICATE_QNAME',                  
      help = 'Used with --why to specify that the predicate shouldnt have its '+
      'magic sets calculated')                    
    op.add_option('--filter', 
                  action='append',
                  default=[],
                  metavar='PATH_OR_URI',
      help = 'The Notation 3 documents to use as a filter (entailments do not particpate in network)')
    op.add_option('--ruleFacts', 
                  action='store_true',
                  default=False,
      help = "Determines whether or not to attempt to parse initial facts from "+
      "the rule graph.  The default is %default")
    op.add_option('--builtins', 
                  default=False,
                  metavar='PATH_TO_PYTHON_MODULE',
      help = "The path to a python module with function definitions (and a "+
      "dicitonary called ADDITIONAL_FILTERS) to use for builtins implementations")    
    op.add_option('--dlp', 
                  action='store_true',
                  default=False,
      help = 'Use Description Logic Programming (DLP) to extract rules from OWL/RDF.  The default is %default')
    op.add_option('--sparqlEndpoint', 
                action='store_true',
                default=False,
    help = 'Indicates that the sole argument is the URI of a SPARQL endpoint to query')

    op.add_option('--ontology', 
                  action='append',
                  default=[],
                  metavar='PATH_OR_URI',
      help = 'The path to an OWL RDF/XML graph to use DLP to extract rules from '+
      '(other wise, fact graph(s) are used)  ') 
      
    op.add_option('--ontologyFormat', 
                default='xml',
                dest='ontologyFormat',
                metavar='RDF_FORMAT',
                choices = ['xml', 'trix', 'n3', 'nt', 'rdfa'],
    help = "The format of the OWL RDF/XML graph specified via --ontology.  The default is %default")
   
    op.add_option('--builtinTemplates', 
                  default=None,
                  metavar='N3_DOC_PATH_OR_URI',
      help = 'The path to an N3 document associating SPARQL FILTER templates to '+
      'rule builtins')        
    op.add_option('--negation', 
                  action='store_true',
                  default=False,                
      help = 'Extract negative rules?')    
    op.add_option('--normalForm', 
                  action='store_true',
                  default=False,                
      help = 'Whether or not to reduce DL axioms & LP rules to a normal form')        
    (options, facts) = op.parse_args()
    
    nsBinds = {'iw':'http://inferenceweb.stanford.edu/2004/07/iw.owl#'}
    for nsBind in options.ns:
        pref,nsUri = nsBind.split('=')
        nsBinds[pref]=nsUri
    
    namespace_manager = NamespaceManager(Graph())
    if options.sparqlEndpoint:
        factGraph = Graph(plugin.get('SPARQL',Store)(facts[0])) 
        options.hybrid = False
    else:
        factGraph = Graph() 
    ruleSet = Ruleset()

    for fileN in options.rules:
        if options.ruleFacts and not options.sparqlEndpoint:
            factGraph.parse(fileN,format='n3')
            print >>sys.stderr,"Parsing RDF facts from ", fileN
        if options.builtins:
            import imp
            userFuncs = imp.load_source('builtins', options.builtins)
            rs = HornFromN3(fileN,
                            additionalBuiltins=userFuncs.ADDITIONAL_FILTERS)
        else:
            rs = HornFromN3(fileN)
        nsBinds.update(rs.nsMapping)
        ruleSet.formulae.extend(rs)
        #ruleGraph.parse(fileN,format='n3')

    ruleSet.nsMapping = nsBinds

    for prefix,uri in nsBinds.items():
        namespace_manager.bind(prefix, uri, override=False)
    closureDeltaGraph = Graph()
    closureDeltaGraph.namespace_manager = namespace_manager
    factGraph.namespace_manager = namespace_manager

    if not options.sparqlEndpoint:
        for fileN in facts:
            factGraph.parse(fileN,format=options.inputFormat)
        
    if not options.sparqlEndpoint and facts:
        for pref,uri in factGraph.namespaces():
            nsBinds[pref]=uri
        
    if options.stdin:
        assert not options.sparqlEndpoint,"Cannot use --stdin with --sparqlEndpoint"
        factGraph.parse(sys.stdin,format=options.inputFormat)
        
    #Normalize namespace mappings
    #prune redundant, rdflib-allocated namespace prefix mappings
    newNsMgr = NamespaceManager(factGraph)
    from FuXi.Rete.Util import CollapseDictionary    
    for k,v in CollapseDictionary(dict([(k,v) 
                                    for k,v in factGraph.namespaces()])).items():
        newNsMgr.bind(k,v)
    factGraph.namespace_manager = newNsMgr
        
    if options.normalForm:
        NormalFormReduction(factGraph)

    if not options.sparqlEndpoint:
        workingMemory = generateTokenSet(factGraph)
    if options.builtins:
        import imp
        userFuncs = imp.load_source('builtins', options.builtins)
        rule_store, rule_graph, network = SetupRuleStore(
                             makeNetwork=True,
                             additionalBuiltins=userFuncs.ADDITIONAL_FILTERS)
    else:
        rule_store, rule_graph, network = SetupRuleStore(makeNetwork=True)
    network.inferredFacts = closureDeltaGraph
    network.nsMap = nsBinds
    
    if options.dlp:
        from FuXi.DLP.DLNormalization import NormalFormReduction
        if options.ontology:
            ontGraph = Graph()
            for fileN in options.ontology:
                ontGraph.parse(fileN,format=options.ontologyFormat)
                for prefix,uri in ontGraph.namespaces():
                    nsBinds[prefix] = uri
                    namespace_manager.bind(prefix, uri, override=False)
                    if options.sparqlEndpoint:
                        factGraph.store.bind(prefix,uri)
        else:
            ontGraph=factGraph
        NormalFormReduction(ontGraph)
        dlp=network.setupDescriptionLogicProgramming(
                                 ontGraph,
                                 addPDSemantics=options.pDSemantics,
                                 constructNetwork=False,
                                 ignoreNegativeStratus=options.negation,
                                 safety = safetyNameMap[options.safety]) 
        ruleSet.formulae.extend(dlp)
    if options.output == 'rif' and not options.why:
        for rule in ruleSet:
            print rule
        if options.negation:
            for nRule in network.negRules:
                print nRule
        
    elif options.output == 'man-owl':
        cGraph = network.closureGraph(factGraph,readOnly=False)
        cGraph.namespace_manager = namespace_manager
        Individual.factoryGraph = cGraph
        if options.classes:
            mapping = dict(namespace_manager.namespaces())
            for c in options.classes:
                pref,uri=c.split(':')
                print Class(URIRef(mapping[pref]+uri)).__repr__(True)
        elif options.properties:
            mapping = dict(namespace_manager.namespaces())
            for p in options.properties:
                pref,uri=p.split(':')
                print Property(URIRef(mapping[pref]+uri))
        else:
            for p in AllProperties(cGraph):
                print p.identifier, first(p.label)
                print repr(p)
            for c in AllClasses(cGraph):
                if options.normalize:
                    if c.isPrimitive():
                        primAnc = [sc for sc in c.subClassOf if sc.isPrimitive()] 
                        if len(primAnc)>1:
                            warnings.warn("Branches of primitive skeleton taxonomy"+
                              " should form trees: %s has %s primitive parents: %s"%(
                             c.qname,len(primAnc),primAnc),UserWarning,1)
                        children = [desc for desc in c.subSumpteeIds()]
                        for child in children:
                            for otherChild in [o for o in children if o is not child]:
                                if not otherChild in [c.identifier 
                                          for c in Class(child).disjointWith]:# and\
                                    warnings.warn("Primitive children (of %s) "+
                                          "must be mutually disjoint: %s and %s"%(
                                      c.qname,
                                      Class(child).qname,
                                      Class(otherChild).qname),UserWarning,1)
#                if not isinstance(c.identifier,BNode):
                print c.__repr__(True)
                    
    if not options.why:
        #Niave construction of graph
        for rule in ruleSet:
            network.buildNetworkFromClause(rule)

    magicSeeds=[]
    if options.why:
        builtinTemplateGraph = Graph()
        if options.builtinTemplates:
            builtinTemplateGraph = Graph().parse(options.builtinTemplates,
                                                format='n3')
        factGraph.templateMap = \
            dict([(pred,template)
                      for pred,_ignore,template in 
                            builtinTemplateGraph.triples(
                                (None,
                                 TEMPLATES.filterTemplate,
                                 None))])
        goals=[]
        if options.why:
            query = ParseSPARQL(options.why) 
            network.nsMap['pml'] = PML
            network.nsMap['gmp'] = GMP_NS
            network.nsMap['owl'] = OWL_NS        
            nsBinds.update(network.nsMap)
            network.nsMap = nsBinds
            if not query.prolog:
                    query.prolog = Prolog(None, [])
                    query.prolog.prefixBindings.update(nsBinds)
            else:
                for prefix, nsInst in nsBinds.items():
                    if prefix not in query.prolog.prefixBindings:
                        query.prolog.prefixBindings[prefix] = nsInst
            goals.extend([(s,p,o) for s,p,o,c in ReduceGraphPattern(
                                        query.query.whereClause.parsedGraphPattern,
                                        query.prolog).patterns])
        dPreds=[]# p for s,p,o in goals ]
        magicRuleNo = 0
        bottomUpDerivedPreds = []
        topDownDerivedPreds  = []
        defaultBasePreds     = []
        defaultDerivedPreds  = []
        mapping = dict(newNsMgr.namespaces())
        for edb in options.edb: 
            pref,uri=edb.split(':')
            defaultBasePreds.append(URIRef(mapping[pref]+uri))
        noMagic = []
        for pred in options.noMagic:
            pref,uri=pred.split(':')
            noMagic.append(URIRef(mapping[pref]+uri))
        if options.ddlGraph:
            ddlGraph = Graph().parse(options.ddlGraph,format='n3')
            defaultDerivedPreds=IdentifyDerivedPredicates(
                                    ddlGraph,
                                    Graph(),
                                    ruleSet)
        else:
            for idb in options.idb: 
                pref,uri=idb.split(':')
                defaultDerivedPreds.append(URIRef(mapping[pref]+uri))
            defaultDerivedPreds.extend(set([p == RDF.type and o or p for s,p,o in goals]))
        
        if options.method == 'gms':
            for goal in goals:
                goalSeed=AdornLiteral(goal).makeMagicPred()
                print >>sys.stderr,"Magic seed fact (used in bottom-up evaluation)",goalSeed
                magicSeeds.append(goalSeed.toRDFTuple())
            if noMagic:
                print >>sys.stderr,"Predicates whose magic sets will not be calculated"
                for p in noMagic:
                    print >>sys.stderr,"\t", factGraph.qname(p)
            for rule in MagicSetTransformation(
                                       factGraph,
                                       ruleSet,
                                       goals,
                                       derivedPreds=bottomUpDerivedPreds,
                                       strictCheck=nameMap[options.strictness],
                                       defaultPredicates=(defaultBasePreds,
                                                          defaultDerivedPreds),
                                       noMagic=noMagic):
                magicRuleNo+=1
                network.buildNetworkFromClause(rule)
            if len(list(ruleSet)):
                print >>sys.stderr,"reduction in size of program: %s (%s -> %s clauses)"%(
                                           100-(float(magicRuleNo)/float(len(list(ruleSet))))*100,
                                           len(list(ruleSet)),
                                           magicRuleNo)
            start = time.time()
            network.feedFactsToAdd(generateTokenSet(magicSeeds))           
            sTime = time.time() - start
            if sTime > 1:
                sTimeStr = "%s seconds"%sTime
            else:
                sTime = sTime * 1000
                sTimeStr = "%s milli seconds"%sTime
            print >>sys.stderr, "Time to calculate closure on working memory: ",sTimeStr                                 
                                                 
            if options.output == 'rif':
                print >>sys.stderr,"Rules used for bottom-up evaluation"
                if network.rules:
                    for clause in network.rules:
                        print >>sys.stderr,clause                    
                else:
                    for clause in factGraph.adornedProgram:
                        print >>sys.stderr,clause 
            if options.output == 'conflict':
                network.reportConflictSet()
            
        elif options.method == 'niave':
            start = time.time()                  
            network.feedFactsToAdd(workingMemory)
            sTime = time.time() - start
            if sTime > 1:
                sTimeStr = "%s seconds"%sTime
            else:
                sTime = sTime * 1000
                sTimeStr = "%s milli seconds"%sTime
            print >>sys.stderr,"Time to calculate closure on working memory: ",sTimeStr
            print >>sys.stderr, network
            if options.output == 'conflict':
                network.reportConflictSet()            
            
        elif options.method in ['sld','bfp']:
            reasoningAlg = TOP_DOWN_METHOD if options.method == 'sld' \
                           else BFP_METHOD
            topDownDPreds = defaultDerivedPreds  if options.ddlGraph else None
            topDownStore=TopDownSPARQLEntailingStore(
                            factGraph.store,
                            factGraph,
                            idb=ruleSet,
                            DEBUG=options.debug,
                            derivedPredicates = topDownDPreds,
                            nsBindings=network.nsMap,
                            decisionProcedure = reasoningAlg,
                            identifyHybridPredicates = 
                            options.hybrid if options.method == 'bfp' else False)
            targetGraph = Graph(topDownStore)
            for pref,nsUri in network.nsMap.items():
                targetGraph.bind(pref,nsUri)      
            start = time.time()                  
            # queryLiteral = EDBQuery([BuildUnitermFromTuple(goal) for goal in goals],
            #                         targetGraph)
            # query = queryLiteral.asSPARQL()
            # print >>sys.stderr, "Goal to solve ", query
            sTime = time.time() - start
            result = targetGraph.query(options.why,initNs=network.nsMap)
            if result.askAnswer:
                sTime = time.time() - start
                if sTime > 1:
                    sTimeStr = "%s seconds"%sTime
                else:
                    sTime = sTime * 1000
                    sTimeStr = "%s milli seconds"%sTime
                print >>sys.stderr,\
    "Time to reach answer ground goal answer of %s: %s"%(result.askAnswer[0],sTimeStr)
            else:
                for rt in result:
                    sTime = time.time() - start
                    if sTime > 1:
                        sTimeStr = "%s seconds"%sTime
                    else:
                        sTime = sTime * 1000
                        sTimeStr = "%s milli seconds"%sTime
                    if options.firstAnswer:
                        break
                    print >>sys.stderr,\
        "Time to reach answer %s via top-down SPARQL sip strategy: %s"%(rt,sTimeStr)
                                
    for fileN in options.filter:
        for rule in HornFromN3(fileN):
            network.buildFilterNetworkFromClause(rule)

    if options.negation and network.negRules and options.method in ['both',
                                                                    'bottomUp']:
        now=time.time()      
        rt=network.calculateStratifiedModel(factGraph)
        print >>sys.stderr,\
        "Time to calculate stratified, stable model (inferred %s facts): %s"%(
                                    rt,
                                    time.time()-now)                
    if options.filter:
        print >>sys.stderr,"Applying filter to entailed facts"
        network.inferredFacts = network.filteredFacts


    if options.closure:
        cGraph = network.closureGraph(factGraph)
        cGraph.namespace_manager = namespace_manager
        print cGraph.serialize(destination=None, 
                               format=options.output, 
                               base=None)
    elif options.output:
        print network.inferredFacts.serialize(destination=None, 
                                              format=options.output, 
                                              base=None)            
if __name__ == '__main__':
    from hotshot import Profile, stats
#    import pycallgraph
#    pycallgraph.start_trace()
#    main()
#    pycallgraph.make_dot_graph('FuXi-timing.png')
#    sys.exit(1)
    p = Profile('fuxi.profile')
    p.runcall(main)
    p.close()    
    s = stats.load('fuxi.profile')
    s.strip_dirs()
    s.sort_stats('time','cumulative','pcalls')
    s.print_stats(.05)
    s.print_callers(.01)
    s.print_callees(.01)            


#!/usr/local/bin/python
# -*- coding: utf-8 -*-
"""
This section defines Horn rules for RIF Phase 1. The syntax and semantics 
incorporates RIF Positive Conditions defined in Section Positive Conditions
"""
import itertools
from FuXi.Horn.PositiveConditions import *
from FuXi.Horn import DATALOG_SAFETY_NONE,DATALOG_SAFETY_STRICT, DATALOG_SAFETY_LOOSE
from rdflib import Variable, BNode, URIRef, Literal, Namespace,RDF,RDFS
from rdflib.Graph import Graph, ConjunctiveGraph

def NetworkFromN3(n3Source,additionalBuiltins=None):
    """
    Takes an N3 / RDF conjunctive graph and returns a ReteNetwork
    """
    from FuXi.Rete.RuleStore import SetupRuleStore
    rule_store, rule_graph, network = SetupRuleStore(
                         makeNetwork=True,
                         additionalBuiltins=additionalBuiltins)
    if isinstance(n3Source,ConjunctiveGraph):
        for ctx in n3Source.contexts():
            for s,p,o in ctx:
                rule_store.add((s,p,o),ctx)
    else:
        for s,p,o in n3Source:
            rule_store.add((s,p,o),n3Source)
    rule_store._finalize()
    for rule in Ruleset(n3Rules=rule_store.rules,nsMapping=rule_store.nsMgr):
        network.buildNetworkFromClause(rule)
    return network

def HornFromDL(owlGraph, safety = DATALOG_SAFETY_NONE, derivedPreds = [],complSkip = []):
    from FuXi.Rete.RuleStore import SetupRuleStore
    ruleStore,ruleGraph,network = SetupRuleStore(makeNetwork=True)
    return network.setupDescriptionLogicProgramming(
                                 owlGraph,
                                 derivedPreds = derivedPreds,
                                 expanded = complSkip,
                                 addPDSemantics=False,
                                 constructNetwork=False,
                                 safety = safety)

def HornFromN3(n3Source,additionalBuiltins=None):
    from FuXi.Rete.RuleStore import SetupRuleStore, N3RuleStore
    if isinstance(n3Source,ConjunctiveGraph):
        store=N3RuleStore(additionalBuiltins=additionalBuiltins)
        for ctx in n3Source.contexts():
            for s,p,o in ctx:
                store.add((s,p,o),ctx)
    else:
        store,graph=SetupRuleStore(n3Source,additionalBuiltins=additionalBuiltins)
    store._finalize()
    return Ruleset(n3Rules=store.rules,nsMapping=store.nsMgr)

def extractVariables(term,existential=True):
    if isinstance(term,existential and BNode or Variable):
        yield term
    elif isinstance(term,Uniterm):
        for t in term.toRDFTuple():
            if isinstance(t,existential and BNode or Variable):
                yield t
                
def iterCondition(condition):
    return isinstance(condition,SetOperator) and condition or iter([condition])
                
class Ruleset(object):
    """
    Ruleset ::= RULE*
    """
    def __init__(self,formulae=None,n3Rules=None,nsMapping=None):
        from FuXi.Rete.RuleStore import N3Builtin
        self.nsMapping = nsMapping and nsMapping or {}        
        self.formulae = formulae and formulae or []
        if n3Rules:
            from FuXi.DLP import breadth_first
            #Convert a N3 abstract model (parsed from N3) into a RIF BLD 
            for lhs,rhs in n3Rules:
                allVars = set()
                for ruleCondition in [lhs,rhs]:
                    for stmt in ruleCondition:
                        if isinstance(stmt,N3Builtin):
                            ExternalFunction(stmt,newNss=self.nsMapping)
#                            print stmt;raise
                        allVars.update([term for term in stmt if isinstance(term,(BNode,Variable))])
                body = [isinstance(term,N3Builtin) and term or
                         Uniterm(list(term)[1],[list(term)[0],list(term)[-1]],
                                 newNss=nsMapping) for term in lhs]
                body = len(body) == 1 and body[0] or And(body)
                head = [Uniterm(p,[s,o],newNss=nsMapping) for s,p,o in rhs]
                head = len(head) == 1 and head[0] or And(head)
                
                #first we identify body variables                        
                bodyVars = set(reduce(lambda x,y:x+y,
                                      [ list(extractVariables(i,existential=False)) 
                                                for i in iterCondition(body) ]))
                #then we identify head variables
                headVars = set(reduce(lambda x,y:x+y,
                                      [ list(extractVariables(i,existential=False)) 
                                                for i in iterCondition(head) ]))
                
                #then we identify those variables that should (or should not) be converted to skolem terms
                updateDict       = dict([(var,BNode()) for var in headVars if var not in bodyVars])
                
                for uniTerm in iterCondition(head):
                    def updateUniterm(uterm):
                        newArg      = [ updateDict.get(i,i) for i in uniTerm.arg ]
                        uniTerm.arg = newArg                    
                    if isinstance(uniTerm,Uniterm):
                        updateUniterm(uniTerm)
                    else:
                        for u in uniTerm:
                            updateUniterm(u)
                                        
                exist=[list(extractVariables(i)) for i in breadth_first(head)]
                e=Exists(formula=head,
                         declare=set(reduce(lambda x,y:x+y,exist,[])))        
                if reduce(lambda x,y:x+y,exist):
                    head=e
                    assert e.declare,exist    
                
                self.formulae.append(Rule(Clause(body,head),declare=allVars))

    def __iter__(self):
        for f in self.formulae:
            yield f

class Rule(object):
    """
    RULE ::= 'Forall' Var* CLAUSE
    
    Example: {?C rdfs:subClassOf ?SC. ?M a ?C} => {?M a ?SC}.
    
    >>> clause = Clause(And([Uniterm(RDFS.subClassOf,[Variable('C'),Variable('SC')]),
    ...                      Uniterm(RDF.type,[Variable('M'),Variable('C')])]),
    ...                 Uniterm(RDF.type,[Variable('M'),Variable('SC')]))
    >>> Rule(clause,[Variable('M'),Variable('SC'),Variable('C')])
    Forall ?M ?SC ?C ( ?SC(?M) :- And( rdfs:subClassOf(?C ?SC) ?C(?M) ) )
    
    """
    def __init__(self,clause,declare=None,nsMapping=None,negativeStratus=False):
        self.negativeStratus=negativeStratus
        self.nsMapping = nsMapping and nsMapping or {}
        self.formula = clause
        self.declare = declare and declare or []
        
    def isSecondOrder(self):
        secondOrder = [pred 
            for pred in itertools.chain(iterCondition(self.formula.head),
                                        iterCondition(self.formula.body))
              if pred.isSecondOrder()]
        return bool(secondOrder)
        
    def isSafe(self):
        """
        A RIF-Core rule, r is safe if and only if
        - r is a rule implication, φ :- ψ, and all the variables that occur 
          in φ are safe in ψ, and all the variables that occur in ψ are bound in ψ;
        - or r is a universal rule, Forall v1,...,vn (r'), n ≥ 1, and r' is safe.        
        
        >>> clause1 = Clause(And([Uniterm(RDFS.subClassOf,[Variable('C'),Variable('SC')]),
        ...                      Uniterm(RDF.type,[Variable('M'),Variable('C')])]),
        ...                 Uniterm(RDF.type,[Variable('M'),Variable('SC')]))
        >>> r1 = Rule(clause1,[Variable('M'),Variable('SC'),Variable('C')])
        >>> clause2 = Clause(And([Uniterm(RDFS.subClassOf,[Variable('C'),Variable('SC')])]),
        ...                 Uniterm(RDF.type,[Variable('M'),Variable('SC')]))
        >>> r2 = Rule(clause2,[Variable('M'),Variable('SC'),Variable('C')])        
        >>> r1.isSafe()
        True
        >>> r2.isSafe()
        False
        
        >>> skolemTerm = BNode()
        >>> e = Exists(Uniterm(RDFS.subClassOf,[skolemTerm,Variable('C')]),declare=[skolemTerm])
        >>> r1.formula.head = e
        >>> r1.isSafe()
        False
        """
        from FuXi.Rete.SidewaysInformationPassing import GetArgs, iterCondition
        assert isinstance(self.formula.head,(Exists,Atomic)),\
                          "Safety can only be checked on rules in normal form"
        for var in itertools.ifilter(lambda term:isinstance(term,
                                                            (Variable,BNode)),
                                     GetArgs(self.formula.head)):
            if not self.formula.body.isSafeForVariable(var):
                return False
        for var in itertools.ifilter(
                         lambda term:isinstance(term,(Variable,BNode)),
                         reduce(lambda l,r:l+r,
                                [GetArgs(lit) 
                                 for lit in iterCondition(self.formula.body)])):
            if not self.formula.body.binds(var):
                return False
        return True

    def n3(self):
        """
        Render a rule as N3 (careful to use e:tuple (_: ?X) skolem functions for existentials in the head)

        >>> clause = Clause(And([Uniterm(RDFS.subClassOf,[Variable('C'),Variable('SC')]),
        ...                      Uniterm(RDF.type,[Variable('M'),Variable('C')])]),
        ...                 Uniterm(RDF.type,[Variable('M'),Variable('SC')]))        
        >>> Rule(clause,[Variable('M'),Variable('SC'),Variable('C')]).n3()
        u'{ ?C rdfs:subClassOf ?SC .\\n ?M a ?C } => { ?M a ?SC }'
                
        """
        return u'{ %s } => { %s }'%(self.formula.body.n3(),
                                    self.formula.head.n3())
#        "Forall %s ( %r )"%(' '.join([var.n3() for var in self.declare]),
#                               self.formula)

    def __eq__(self,other):
        return hash(self.formula)==hash(other.formula)
        
    def __hash__(self):
        """
        >>> a=Clause(And([Uniterm(RDFS.subClassOf,[Variable('C'),Variable('SC')]),
        ...             Uniterm(RDF.type,[Variable('M'),Variable('C')])]),
        ...        Uniterm(RDF.type,[Variable('M'),Variable('SC')]))
        >>> b=Clause(And([Uniterm(RDFS.subClassOf,[Variable('C'),Variable('SC')]),
        ...             Uniterm(RDF.type,[Variable('M'),Variable('C')])]),
        ...        Uniterm(RDF.type,[Variable('M'),Variable('SC')]))
        >>> d=set()
        >>> d.add(a)
        >>> b in d
        True
        >>> hash(a) == hash(b)
        True
        >>> EX_NS = Namespace('http://example.com/')
        >>> a=Clause(Uniterm(RDF.type,[Variable('C'),EX_NS.Foo]),
        ...          Uniterm(RDF.type,[Variable('C'),EX_NS.Bar]))
        >>> b=Clause(Uniterm(RDF.type,[Variable('C'),EX_NS.Bar]),
        ...          Uniterm(RDF.type,[Variable('C'),EX_NS.Foo]))
        >>> a == b
        False
        """
        return hash(self.formula)

    def __repr__(self):
        return "Forall %s ( %r )"%(' '.join([var.n3() for var in self.declare]),
                               self.formula)    
        
def NormalizeBody(rule):
    from FuXi.Rete.RuleStore import N3Builtin
    from itertools import groupby, chain
    newBody  = []
    builtIns = []
    if isinstance(rule.formula.body,And):
        for lit in rule.formula.body:
            if isinstance(lit,N3Builtin):
                builtIns.append(lit)
            else:
                newBody.append(lit)
        newBody.extend(builtIns)
        rule.formula.body = And(newBody)
    return rule
        
class Clause(object):
    """
    Facts are *not* modelled formally as rules with empty bodies
    
    Implies ::= ATOMIC ':-' CONDITION
    
    Use body / head instead of if/then (native language clash)
    
    Example: {?C rdfs:subClassOf ?SC. ?M a ?C} => {?M a ?SC}.
    
    >>> Clause(And([Uniterm(RDFS.subClassOf,[Variable('C'),Variable('SC')]),
    ...             Uniterm(RDF.type,[Variable('M'),Variable('C')])]),
    ...        Uniterm(RDF.type,[Variable('M'),Variable('SC')]))
    ?SC(?M) :- And( rdfs:subClassOf(?C ?SC) ?C(?M) )
    """
    def __init__(self,body,head):
        self.body = body
        self.head = head
        from FuXi.Rete.Network import HashablePatternList
        antHash=HashablePatternList(
                    [term.toRDFTuple() for term in body],skipBNodes=True)
        consHash=HashablePatternList(
                    [term.toRDFTuple() for term in head],skipBNodes=True)
        self._bodyHash = hash(antHash)
        self._headHash = hash(consHash)             
        self._hash     = hash((self._headHash,self._bodyHash))                                                      
        
    def __eq__(self,other):
        return hash(self)==hash(other)
        
    def __hash__(self):
        """
        >>> a=Clause(And([Uniterm(RDFS.subClassOf,[Variable('C'),Variable('SC')]),
        ...             Uniterm(RDF.type,[Variable('M'),Variable('C')])]),
        ...        Uniterm(RDF.type,[Variable('M'),Variable('SC')]))
        >>> b=Clause(And([Uniterm(RDFS.subClassOf,[Variable('C'),Variable('SC')]),
        ...             Uniterm(RDF.type,[Variable('M'),Variable('C')])]),
        ...        Uniterm(RDF.type,[Variable('M'),Variable('SC')]))
        >>> d=set()
        >>> d.add(a)
        >>> b in d
        True
        >>> hash(a) == hash(b)
        True
        
        >>> d={a:True}
        >>> b in d
        True
        """
        return self._hash
        
    def asTuple(self):
        return (self.body,self.head)
        
    def __repr__(self):
        if isinstance(self.body,SetOperator) and not len(self.body):
            return "%r :-"%self.head
        return "%r :- %r"%(self.head,self.body)
    
    def n3(self):
        return u'{ %s } => { %s }'%(self.body.n3(),self.head.n3())    
        
def test():
    import doctest
    doctest.testmod()

if __name__ == '__main__':
    test()


#!/usr/bin/env python
# encoding: utf-8
import sys, unittest
from pprint import pprint
from cStringIO import StringIO
from rdflib.Graph import Graph
from rdflib import Namespace
from FuXi.Rete.RuleStore import SetupRuleStore
from FuXi.Rete.Util import generateTokenSet
from FuXi.DLP.DLNormalization import NormalFormReduction

EX       = Namespace('http://example.org/')
EX_TERMS = Namespace('http://example.org/terms/')

expected_triples = [
    (EX.john,EX_TERMS.has_sibling,EX.jack),
    (EX.john,EX_TERMS.brother,EX.jack),
    (EX.jack,EX_TERMS.has_brother,EX.john),    
]

ABOX=\
"""
@prefix exterms: <http://example.org/terms/> .
@prefix : <http://example.org/> .

:john exterms:has_brother :jack .
:jack exterms:brother     :john .
"""

TBOX=\
"""
@prefix exterms: <http://example.org/terms/> .
@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix owl: <http://www.w3.org/2002/07/owl#>.

exterms:Agent
	a rdfs:Class .

exterms:Person
	a rdfs:Class ;
	rdfs:subClassOf exterms:Agent .

exterms:has_sibling
	a rdf:Property .

exterms:has_brother
	a rdf:Property ;
	rdfs:subPropertyOf exterms:has_sibling ;
	rdfs:domain exterms:Person ;
	rdfs:range exterms:Person .
	
exterms:brother
	a rdf:Property ;
	owl:equivalentProperty exterms:has_brother ;
	rdfs:domain exterms:Person ;
	rdfs:range exterms:Person .
	
"""

class test_superproperty_entailment(unittest.TestCase):
    def setUp(self):
        self.rule_store, self.rule_graph, self.network = SetupRuleStore(makeNetwork=True)
        self.tBoxGraph = Graph().parse(StringIO(TBOX), format='n3')

        self.aBoxGraph = Graph().parse(StringIO(ABOX), format='n3')
        NormalFormReduction(self.tBoxGraph)
    def testReasoning(self):
        print 'setting up DLP...'
        self.network.setupDescriptionLogicProgramming(self.tBoxGraph)
        pprint(list(self.network.rules))
        print self.network
        
        print 'feeding TBox... '
        self.network.feedFactsToAdd(generateTokenSet(self.tBoxGraph))
        print 'feeding ABox...'
        self.network.feedFactsToAdd(generateTokenSet(self.aBoxGraph))

        self.network.inferredFacts.bind('ex',EX)
        self.network.inferredFacts.bind('exterms',EX_TERMS)        
        print self.network.inferredFacts.serialize(format='n3')        

        print 'Checking...'
        for triple in expected_triples:
            self.failUnless(triple in self.network.inferredFacts,"Missing %s"%(repr(triple)))
    
if __name__ == '__main__':
	unittest.main()


import unittest, os, time, sys 
from pprint import pprint
from cStringIO import StringIO 
from rdflib import RDF, URIRef 
from FuXi.Rete import * 
from FuXi.Rete.Util import generateTokenSet
from FuXi.Rete.Magic import MagicSetTransformation, PrettyPrintRule
from FuXi.Rete.RuleStore import N3RuleStore, SetupRuleStore 
from FuXi.Rete.Util import renderNetwork, generateTokenSet 
from FuXi.Horn.PositiveConditions import Uniterm, BuildUnitermFromTuple 
from FuXi.DLP import PrepareHornClauseForRETE
from FuXi.Horn.HornRules import HornFromN3 
from FuXi.Syntax.InfixOWL import *
from cStringIO import StringIO
from rdflib import plugin, Namespace, Variable
from rdflib.store import Store 
from rdflib.Graph import Graph 
N3_PROGRAM=\
""" 
@prefix m: <http://example.com/#>. 
@prefix rdf:  <http://www.w3.org/1999/02/22-rdf-syntax-ns#> . 
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> . 

_:1 a ptrec:Event_morbidity_renal_failure_requiring_dialysis;
  ptrec:hasEventPlace ptrec:EventPlace_ccf_main_campus;
  csqr:hasHospitalization _12:fb8dfdd-d695-4750-9da9-33d2112a9433;
  fztime:startsNoEarlierThan "2010-03-07T15:51:00"^^<http://www.w3.org/2001/XMLSchema#dateTime>;
  fztime:startsNoLaterThan "2010-03-07T15:51:00"^^<http://www.w3.org/2001/XMLSchema#dateTime>. 

""" 
N3_FACTS=\
""" 
@prefix m: <http://example.com/#>. 
@prefix rdf:  <http://www.w3.org/1999/02/22-rdf-syntax-ns#> . 
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> . 
m:Detection a rdfs:Class. 
m:Inference a rdfs:Class. 
:det1 a m:Detection. 
:det1 m:name "Inference1". 
:det2 a m:Detection. 
:det2 m:name "Inference2". 
""" 
class ExistentialInHeadTest(unittest.TestCase): 
    def testExistentials(self): 
        store = plugin.get('IOMemory',Store)() 
        store.open('') 
        ruleStore = N3RuleStore() 
        ruleGraph = Graph(ruleStore) 
        ruleGraph.parse(StringIO(N3_PROGRAM),format='n3') 
        factGraph = Graph(store) 
        factGraph.parse(StringIO(N3_FACTS),format='n3') 
        deltaGraph = Graph(store) 
        network = ReteNetwork(ruleStore, 
                              initialWorkingMemory=generateTokenSet(factGraph), 
                              inferredTarget = deltaGraph) 
        inferenceCount = 0 
        for inferredFact in network.inferredFacts.subjects(
                                       predicate=RDF.type, 
                                       object=URIRef('http://example.com/#Inference')): 
            inferenceCount = inferenceCount + 1
        print network.inferredFacts.serialize(format='n3') 
        self.failUnless(inferenceCount > 1,  'Each rule firing should introduce a new BNode!') 
        cg = network.closureGraph(factGraph, store=ruleStore) 
        #print cg.serialize(format="n3") 

FACTS=\
""" 
@prefix ex: <http://example.com/#>. 
@prefix rdf:  <http://www.w3.org/1999/02/22-rdf-syntax-ns#> . 
@prefix cpr:  <http://purl.org/cpr/> . 
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> . 
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .

ex:evt1 a ex:RenalFailureRequiringDialysis;
    cpr:startsNoEarlierThan "2010-03-07T15:51:00"^^xsd:dateTime .

ex:evt2 a cpr:therapeutic-procedure, ex:QualifyingOperation .
ex:evt2 cpr:startsNoEarlierThan "2010-03-05T10:32:00"^^xsd:dateTime .

ex:evt3 ex:intervalContains ex:evt1, ex:evt2 .  
  
""" 

N3_RULES=\
"""
@prefix ex: <http://example.com/#>. 
@prefix rdf:  <http://www.w3.org/1999/02/22-rdf-syntax-ns#> . 
@prefix cpr:  <http://purl.org/cpr/> . 
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> . 
@prefix str:    <http://www.w3.org/2000/10/swap/string#>.

{ ?IDX_OP a ex:QualifyingOperation;
          cpr:startsNoEarlierThan ?IDX_OP_START .
 ?HOSP  ex:intervalContains  ?IDX_OP, ?EVENT .
 ?EVENT a cpr:therapeutic-procedure ;
        cpr:startsNoEarlierThan ?EVT_START .
 ?EVT_START str:greaterThan ?IDX_OP_START  } => { ?EVENT a ex:PostOpInHospitalEvent } .
"""

EX_NS  = Namespace('http://example.com/#')
EX     = ClassNamespaceFactory(EX_NS)
CPR_NS = Namespace('http://purl.org/cpr/')
CPR    = ClassNamespaceFactory(CPR_NS)             
             
derivedPredicate = [EX_NS.PostOpInHospitalEvent,EX_NS.RenalFailure]                         
             
class SkolemMachine(unittest.TestCase):
    def setUp(self):
        ruleStore,ruleGraph,network=SetupRuleStore(makeNetwork=True)
        self.network= network
        self.factGraph = Graph().parse(StringIO(SKOLEM_MACHINE_FACTS),format='n3')
        for rule in HornFromN3(StringIO(SKOLEM_MACHINE_RULES)):
            self.network.buildNetworkFromClause(rule)
        self.network.feedFactsToAdd(generateTokenSet(self.factGraph))
            
if __name__ == "__main__": 
    ontGraph = Graph()
    ontGraph.bind('ex' , EX_NS)
    ontGraph.bind('cpr', CPR_NS)
    Individual.factoryGraph = ontGraph
    renalFailure = EX.RenalFailure
    conjunct = EX.PostOpInHospitalEvent & EX.RenalFailureRequiringDialysis
    renalFailure += conjunct
    rules = set()
    rules.update(HornFromN3(StringIO(N3_RULES)))
    store, graph, network = SetupRuleStore(StringIO(N3_RULES),makeNetwork=True)
    rules.update(network.setupDescriptionLogicProgramming(
                    ontGraph,
                    derivedPreds=derivedPredicate,
                    addPDSemantics=False,
                    classifyTBox=False))
    pprint(list(rules))
    varX = Variable('x')
    goals = [(varX,RDF.type,_class) for _class in derivedPredicate]
    for rule in MagicSetTransformation(None,
                                       rules,
                                       goals,
                                       derivedPreds=derivedPredicate):
        PrepareHornClauseForRETE(rule)
        network.buildNetworkFromClause(rule)  
        PrettyPrintRule(rule)
    facts = Graph().parse(StringIO(FACTS),format='n3')
    network.feedFactsToAdd(generateTokenSet(facts))          
    print network.inferredFacts.serialize(format='turtle')
    print facts.serialize(format='turtle')
    
    # unittest.main() 


import copy
from itertools import chain
from FuXi.Horn.PositiveConditions import QNameManager,SetOperator, Condition, Or, And
from FuXi.Rete.Util import selective_memoize
from FuXi.Rete.RuleStore import *
from FuXi.Rete.Proof import ImmutableDict
from rdflib import URIRef, RDF, Namespace, Variable, Literal
from rdflib.util import first
from FuXi.Rete.BetaNode import project
from FuXi.Rete.SidewaysInformationPassing import GetArgs, iterCondition, GetOp, GetVariables

def normalizeBindingsAndQuery(vars,bindings,conjunct):
    """
    Takes a query in the form of a list of variables to bind to
    an a priori set of bindings and a conjunct of literals and applies the bindings
    returning:
     - The remaining variables that were not substituted
     - The (possibly grounded) conjunct of literals
     - The bindings minus mappings involving substituted variables 
    
    """
    _vars = set(vars)
    bindingDomain = set(bindings.keys())
    appliedBindings = False
    if bindings:
        #Apply a priori substitutions
        for lit in conjunct:
            substitutedVars = bindingDomain.intersection(lit.toRDFTuple())
            lit.ground(bindings)
            if substitutedVars:
                appliedBindings = True
                _vars.difference_update(substitutedVars)
    return list(_vars),conjunct, \
           project(bindings,_vars,inverse=True) if appliedBindings else bindings

def tripleToTriplePattern(graph,term):
   if isinstance(term,N3Builtin):
       template = graph.templateMap[term.uri]
       return "FILTER(%s)"%(template%(term.argument.n3(),
                                      term.result.n3()))
   else:
       return "%s %s %s"%tuple([renderTerm(graph,term) 
                                   for term in term.toRDFTuple()])

@selective_memoize([0])
def normalizeUri(rdfTerm,revNsMap):
   """
   Takes an RDF Term and 'normalizes' it into a QName (using the registered prefix)
   or (unlike compute_qname) the Notation 3 form for URIs: <...URI...> 
   """
   try:
       namespace, name = split_uri(rdfTerm)
       namespace = URIRef(namespace)
   except:
       if isinstance(rdfTerm,Variable):
           return "?%s"%rdfTerm
       else:
           return "<%s>"%rdfTerm
   prefix = revNsMap.get(namespace)
   if prefix is None and isinstance(rdfTerm,Variable):
       return "?%s"%rdfTerm
   elif prefix is None:
       return "<%s>"%rdfTerm
   else:
       qNameParts = compute_qname(rdfTerm,revNsMap)         
       return ':'.join([qNameParts[0],qNameParts[-1]])    

@selective_memoize([0])
def compute_qname(uri,revNsMap):
   namespace, name = split_uri(uri)
   namespace = URIRef(namespace)
   prefix = revNsMap.get(namespace)
   if prefix is None:
       prefix = "_%s" % len(revNsMap)
       revNsMap[namespace]=prefix
   return (prefix, namespace, name)

def renderTerm(graph,term):
   if term == RDF.type:
       return ' a '
   elif isinstance(term,URIRef):
       qname = normalizeUri(term,hasattr(graph,'revNsMap') and graph.revNsMap or \
                            dict([(u,p) for p,u in graph.namespaces()]))
       return qname[0] == '_' and u"<%s>"%term or qname
   elif isinstance(term,Literal):
       return term.n3()
   else:
       try:
           return isinstance(term,BNode) and term.n3() or graph.qname(term)
       except:
           return term.n3()


def RDFTuplesToSPARQL(conjunct, 
                     edb, 
                     isGround=False, 
                     vars=[],
                     symmAtomicInclusion=False):
   """
   Takes a conjunction of Horn literals and returns the 
   corresponding SPARQL query 
   """
   queryType = isGround and "ASK" or "SELECT %s"%(' '.join([v.n3() 
                                                            for v in vars]))
   queryShell = len(conjunct)>1 and "%s {\n%s\n}" or "%s { %s }"

   if symmAtomicInclusion:
       if vars:
           var = vars.pop()
           prefix = "%s a ?KIND"%var.n3()
       else:

           prefix = "%s a ?KIND"%first([first(iterCondition(lit)).arg[0].n3() for lit in conjunct])
       conjunct = ( i.formulae[0] if isinstance(i,And) else i for i in conjunct )
       subquery = queryShell%(queryType,
                              "%s\nFILTER(%s)"%(
                            prefix,
                            ' ||\n'.join([
                              '?KIND = %s'%edb.qname(GetOp(lit)) 
                                   for lit in conjunct])))        
   else: 
       subquery = queryShell%(queryType,' .\n'.join(['\t'+tripleToTriplePattern(
                                                             edb,
                                                             lit) 
                                 for lit in conjunct ]))
   return subquery

#@selective_memoize([0,1],['vars','symmAtomicInclusion'])
def RunQuery(subQueryJoin,
            bindings,
            factGraph,
            vars=None,
            debug = False,
            symmAtomicInclusion = False):
   initialNs = hasattr(factGraph,'nsMap') and factGraph.nsMap or \
               dict([(k,v) for k,v in factGraph.namespaces()])

   if not subQueryJoin:
       return False
   if not vars:
       vars=[]
   if bool(bindings):
       #Apply a priori substitutions
       openVars,conjGroundLiterals,bindings  = \
               normalizeBindingsAndQuery(set(vars),
                                         bindings,
                                         subQueryJoin)
       vars=list(openVars)
   else:
       conjGroundLiterals = subQueryJoin
   isGround = not vars
   subquery = RDFTuplesToSPARQL(conjGroundLiterals,
                                factGraph,
                                isGround,
                                [v for v in vars],
                                symmAtomicInclusion)
   rt = factGraph.query(subquery,
                        initNs = initialNs)#,
                        #DEBUG = debug)
   projectedBindings = vars and project(bindings,vars) or bindings
   if isGround:
       if debug:
           print >>sys.stderr, "%s%s-> %s"%(
                         subquery,
                         projectedBindings and 
                         " %s apriori binding(s)"%len(projectedBindings) or '',
                         rt.askAnswer[0])
       return subquery,rt.askAnswer[0]
   else:
       rt = len(vars)>1 and ( dict([(vars[idx],i) 
                                      for idx,i in enumerate(v)]) 
                                           for v in rt ) \
              or ( dict([(vars[0],v)]) for v in rt )
       if debug:
           print >>sys.stderr, "%s%s-> %s"%(
                   subquery,
                   projectedBindings and 
                   " %s apriori binding(s)"%len(projectedBindings) or '',                                
                   rt and '[]')# .. %s answers .. ]'%len(rt) or '[]')
       return subquery,rt

class EDBQuery(QNameManager,SetOperator,Condition):
    """
    A list of frames (comprised of EDB predicates) meant for evaluation over a large EDB
    
    lst is a conjunct of terms
    factGraph is the RDF graph to evaluate queries over
    returnVars is the return variables (None, the default, will cause the list
     to be built via instrospection on lst)
    bindings is a solution mapping to apply to the terms in lst 
    
    
    """
    def __init__(self, 
                 lst, 
                 factGraph,                  
                 returnVars=None, 
                 bindings={}, 
                 varMap={}, 
                 symIncAxMap = {}, 
                 symmAtomicInclusion = False):
        self.factGraph           = factGraph
        self.varMap              = varMap
        self.symmAtomicInclusion = symmAtomicInclusion
        self.formulae            = lst
        self.naf                 = False

        #apply an apriori solutions
        if bool(bindings):
            #Apply a priori substitutions
            openVars,termList,bindings  = \
                    normalizeBindingsAndQuery(set(returnVars) 
                        if returnVars else [v for v in self.getOpenVars()],
                                              bindings,
                                              lst)
            self.returnVars = list(openVars)
        else:
            if returnVars is None:
                #return vars not specified, but meant to be determined by 
                #constructor 
                self.returnVars = self.getOpenVars()
            else:
                #Note if returnVars is an empty list, this
                self.returnVars = (returnVars if isinstance(returnVars,list) 
                                      else list(returnVars)) if returnVars else []
            termList = lst
            
        super(EDBQuery, self).__init__(termList)
        self.bindings            = bindings.normalize() \
                                        if isinstance(
                                            bindings,
                                            ImmutableDict) else bindings

    def copy(self):
        """
        A shallow copy of an EDB query
        """
        return EDBQuery([copy.deepcopy(t) for t in self.formulae],
                        self.factGraph,
                        self.returnVars,
                        self.bindings.copy(),
                        self.varMap.copy(),
                        symmAtomicInclusion = self.symmAtomicInclusion)
        
    def renameVariables(self, varMap):
        for item in self.formulae:
            item.renameVariables(varMap)
        
    def ground(self,mapping):
        appliedVars = set()
        for item in self.formulae:
            if isinstance(item,Or):
                for _item in item.formulae:
                    appliedVars.update(item.ground(mapping))
            else:
                appliedVars.update(item.ground(mapping))
        self.bindings = project(self.bindings,appliedVars,True)
        return appliedVars
                
    def accumulateBindings(self, bindings):
        """
        """
        self.bindings.update(project(bindings,self.getOpenVars(),inverse=True))

    def getOpenVars(self):
        return list(
                 set(
                   reduce(
                     lambda x,y:x+y,
                     map(lambda arg:list(GetVariables(arg,secondOrder=True)),
                         self.formulae))))

    def applyMGU(self,substitutions):
        for term in self.formulae:
            term.renameVariables(substitutions)
        self.bindings = dict([(substitutions.get(k,k),v) 
                            for k,v in self.bindings.items()])

    def evaluate(self,debug = False, symmAtomicInclusion = False):
        return     RunQuery(self.formulae,
                            self.bindings,
                            self.factGraph,
                            vars=self.returnVars,
                            debug = debug,
                            symmAtomicInclusion = symmAtomicInclusion)
        
    def asSPARQL(self):
        initialNs = hasattr(self.factGraph,'nsMap') and self.factGraph.nsMap or \
                    dict([(k,v) for k,v in self.factGraph.namespaces()])
        return RDFTuplesToSPARQL(self.formulae,
                                 self.factGraph,
                                 not self.returnVars,
                                 self.returnVars,
                                 self.symmAtomicInclusion)
        
    def __hash__(self):
        from FuXi.Rete.Network import HashablePatternList
        conj=HashablePatternList(
                    [term.toRDFTuple() for term in self.formulae],
                    skipBNodes=True)
        return hash(conj) 
        
    def extend(self, query, newVarMap = None):
        assert not query.symmAtomicInclusion  
        assert not self.symmAtomicInclusion  
        if newVarMap:
            query.renameVariables(newVarMap)
            self.varMap.update(newVarMap)
        self.formulae.extend([term for term in query.formulae 
                                if term not in self.formulae])
        self.bindings.update(query.bindings)
        
    def __repr__(self):
        return "EDBQuery(%s%s)"%(self.repr(self.symmAtomicInclusion and 'Or' or 'And'),
                       self.bindings and ',%s'%self.bindings or '')
        
def test():
    import doctest
    doctest.testmod()

if __name__ == '__main__':
    test()


#!/usr/bin/env python
# encoding: utf-8
import unittest
from pprint import pprint
from rdflib import Variable, Namespace, Literal
from rdflib.Graph import Graph
from cStringIO import StringIO
from FuXi.Rete.RuleStore import SetupRuleStore
from FuXi.Syntax.InfixOWL import OWL_NS
from FuXi.Horn.HornRules import HornFromN3
from FuXi.Horn.PositiveConditions import BuildUnitermFromTuple
from FuXi.Rete.TopDown import PrepareSipCollection, SipStrategy
from FuXi.Rete.Magic import SetupDDLAndAdornProgram
from FuXi.Rete.SidewaysInformationPassing import SIPRepresentation

EX = Namespace('http://example.org/')

FACTS=\
"""
@prefix ex: <http://example.org/> .
@prefix owl: <http://www.w3.org/2002/07/owl#>.

ex:foo ex:x "xxxx";
       owl:sameAs ex:bar .
ex:bar ex:y "yyyy" .
"""

RULES=\
"""
@prefix owl: <http://www.w3.org/2002/07/owl#>.

{ ?x owl:sameAs ?y } => { ?y owl:sameAs ?x } .
{ ?x owl:sameAs ?y . ?x ?p ?o } => { ?y ?p ?o } .
"""

GOALS = [
    (EX.foo,EX.y,Variable('o')),
]

class test_sameAs(unittest.TestCase):
    def setUp(self):
        self.rule_store, self.rule_graph, self.network = SetupRuleStore(makeNetwork=True)
        self.graph = Graph().parse(StringIO(FACTS), format='n3')
        adornedProgram = SetupDDLAndAdornProgram(
                                       self.graph,
                                       HornFromN3(StringIO(RULES)),
                                       GOALS,
                                       derivedPreds=[OWL_NS.sameAs])
        pprint(list(self.graph.adornedProgram))
    def testSmushing(self):
        sipCollection = PrepareSipCollection(self.graph.adornedProgram)
        print self.graph.serialize(format='n3')
        for arc in SIPRepresentation(sipCollection):
            print arc
        success = False
        for goal in GOALS:
            goalLiteral = BuildUnitermFromTuple(goal)
            print "Query / goal: ", goalLiteral
            for ans,node in SipStrategy(
                        goal,
                        sipCollection,
                        self.graph,
                        [OWL_NS.sameAs],
                        debug = False):
                if ans[Variable('o')] == Literal('yyyy'):
                    success = True
                    print "Found solution!", ans
                    break
        self.failUnless(success,"Unable to proove %s"%(repr((EX.foo,EX.y,Literal('yyyy')))))

if __name__ == '__main__':
	unittest.main()